# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""An example that computes the Nash bargaining score from negotiations.

This uses the bargaining game that was introduced in:

[1] Lewis et al., Deal or no deal? End-to-end learning of negotiation
    dialogues, 2017. https://arxiv.org/abs/1706.05125
[2] David DeVault, Johnathan Mell, and Jonathan Gratch.
    2015. Toward Natural Turn-taking in a Virtual Human Negotiation Agent

It computes the empirical Nash bargaining score (NBS) from three sources:
  - Human play
  - IS-MCTS in self-play
  - A theoretical maximum NBS if the players had full information and can see
    each other's utilities and then maximize their NBS.

These are all run on a data set extracted from the Lewis et al. '17 data set:
https://github.com/facebookresearch/end-to-end-negotiator/blob/master/src/data/negotiate/data.txt

This example is inspired by the paper (Iwasa and Fujita, "Prediction of Nash
Bargaining Solution in Negotiation Dialogue", 2018).
"""

from absl import app
from absl import flags
import numpy as np

from open_spiel.python import games  # pylint: disable=unused-import
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_string("data_file", None, "Lewis et al. '17 data set file")
flags.DEFINE_string("instances_file", "/tmp/instances.txt",
                    "Filename for the temp instances database file.")


class Instance(object):
  """An instance of a bargaining problem."""

  def __init__(self, pool, p1values, p2values):
    self.pool = np.array(pool)
    self.p1values = np.array(p1values)
    self.p2values = np.array(p2values)
    assert 5 <= sum(pool) <= 7
    assert np.dot(pool, p1values) == 10
    assert np.dot(pool, p2values) == 10

  def __str__(self):
    return (",".join([str(x) for x in self.pool]) + " " +
            ",".join([str(x) for x in self.p1values]) + " " +
            ",".join([str(x) for x in self.p2values]))


class Negotiation(object):
  """An instance of a bargaining game."""

  def __init__(self, instance, outcome, rewards):
    self.instance = instance
    self.outcome = outcome
    self.rewards = rewards

  def __str__(self):
    return (str(self.instance) + " " + str(self.outcome) + " " +
            str(self.rewards))


def dialogue_matches_prev_line(line1, line2):
  """Checks if the dialogue matches the previous line's."""
  parts1 = line1.split(" ")
  parts2 = line2.split(" ")
  for i in range(6, min(len(parts1), len(parts2))):
    if parts1[i] == "YOU:" or parts1[i] == "THEM:":
      if parts1[i] == "YOU:" and parts2[i] != "THEM:":
        return False
      if parts1[i] == "THEM:" and parts2[i] != "YOU:":
        return False
    elif parts1[i] != parts2[i]:
      return False
    if parts1[i] == "<selection>":
      break
  return True


# pylint: disable=line-too-long
def parse_dataset(filename):
  """Parse the Lewis et al. '17 data file."""
  # book, hat, ball
  # Example format
  # 1 0 4 2 1 2 YOU: i would like 4 hats and you can have the rest . <eos> THEM: deal <eos> YOU: <selection> item0=0 item1=4 item2=0 <eos> reward=8 agree 1 4 4 1 1 2
  # 1 4 4 1 1 2 THEM: i would like 4 hats and you can have the rest . <eos> YOU: deal <eos> THEM: <selection> item0=1 item1=0 item2=1 <eos> reward=6 agree 1 0 4 2 1 2
  # 1 6 3 0 2 2 YOU: you can have all the hats if i get the book and basketballs . <eos> THEM: <selection> item0=1 item1=3 item2=2 <eos> reward=10 disagree 1 2 3 2 2 1
  # 1 10 3 0 1 0 YOU: hi i would like the book and ball and you can have the hats <eos> THEM: i can give you either the book or the ball <eos> YOU: ill take the book <eos> THEM: ok i will take the hats and ball <eos> YOU: deal <eos> THEM: <selection> item0=1 item1=0 item2=0 <eos> reward=10 agree 1 2 3 2 1 2
  # 1 2 3 2 1 2 THEM: hi i would like the book and ball and you can have the hats <eos> YOU: i can give you either the book or the ball <eos> THEM: ill take the book <eos> YOU: ok i will take the hats and ball <eos> THEM: deal <eos> YOU: <selection> item0=0 item1=3 item2=1 <eos> reward=8 agree 1 10 3 0 1 0
  contents = pyspiel.read_contents_from_file(filename, "r")
  lines = contents.split("\n")
  cur_nego = None
  negotiations = []
  instances = []

  for line_no in range(len(lines)):
    line = lines[line_no]
    if line:
      parts = line.split(" ")
      # parse the line to add a new negotiation
      pool = [int(parts[0]), int(parts[2]), int(parts[4])]
      my_values = [int(parts[1]), int(parts[3]), int(parts[5])]
      pool2 = [int(parts[-6]), int(parts[-4]), int(parts[-2])]
      other_values = [int(parts[-5]), int(parts[-3]), int(parts[-1])]
      assert pool == pool2
      rewards = [0, 0]
      add_nego = False
      outcome_str = parts[-7]  # this will be "agree" or "disagree"
      if parts[6] == "YOU:":
        player_id = 0
        instance = Instance(pool, my_values, other_values)
      elif parts[6] == "THEM:":
        player_id = 1
        instance = Instance(pool, other_values, my_values)
      else:
        assert False, parts[6]
      outcome = False
      my_reward = 0
      instances.append(instance)
      if "disconnect" in line:
        continue
      # sometimes there is a "no agreement" in the rewards section
      if (outcome_str == "disagree" or
          (parts[-9] + " " + parts[-8]) == "reward=no agreement" or
          parts[-8] == "reward=disconnect"):
        # do not parse the reward, but must still parse the next line
        add_nego = False
      elif outcome_str == "agree":
        outcome = True
        reward_parts = parts[-8].split("=")
        assert len(reward_parts) == 2, f"reward parts str: {parts[-8]}"
        assert reward_parts[0] == "reward"
        my_reward = int(reward_parts[1])
      else:
        assert False, f"Bad outcome: {outcome_str}"
      if cur_nego is None:
        rewards[player_id] = my_reward
        if player_id == 0:
          cur_nego = Negotiation(instance, outcome, rewards)
        else:
          cur_nego = Negotiation(instance, outcome, rewards)
      else:
        # There are some in the data set that are incomplete (i.e. are missing the second perspective).
        # We should not count these.
        if dialogue_matches_prev_line(line, lines[line_no - 1]):
          assert list(cur_nego.instance.pool) == pool
          if player_id == 1:
            assert list(cur_nego.instance.p2values) == my_values
            assert list(cur_nego.instance.p1values) == other_values
          elif player_id == 0:
            assert list(cur_nego.instance.p1values) == my_values
            assert list(cur_nego.instance.p2values) == other_values
          cur_nego.rewards[player_id] = my_reward
          add_nego = True
        else:
          # not matching, treat as new negotiation
          rewards[player_id] = my_reward
          if player_id == 0:
            cur_nego = Negotiation(instance, outcome, rewards)
          else:
            cur_nego = Negotiation(instance, outcome, rewards)
          add_nego = False
      if add_nego or outcome_str == "disagree":
        negotiations.append(cur_nego)
        print(str(cur_nego))
        print(len(negotiations))
        cur_nego = None
        if outcome_str != "disagree":
          # same instance was added twice, so remove the last one
          instances.pop()
  return instances, negotiations


def write_instances_file(negotiations, filename):
  contents = ""
  for nego in negotiations:
    contents += str(nego.instance) + "\n"
  pyspiel.write_contents_to_file(filename, "w", contents)


def compute_nbs_from_simulations(game, num_games, bots):
  """Compute empirical NBS from simulations."""
  sum_nbs = 0
  for _ in range(num_games):
    state = game.new_initial_state()
    while not state.is_terminal():
      if state.is_chance_node():
        # Chance node: sample an outcome
        outcomes = state.chance_outcomes()
        action_list, prob_list = zip(*outcomes)
        action = np.random.choice(action_list, p=prob_list)
        state.apply_action(action)
      else:
        player = state.current_player()
        action = bots[player].step(state)
        state.apply_action(action)
    returns = state.returns()
    sum_nbs += np.prod(returns)
  return sum_nbs / num_games


class MaxBot(object):
  """Finds the single (deterministic) trade offer that maximizes the NBS."""

  def __init__(self):
    pass

  def step(self, state):
    """Returns the NBS-maximizing action.

    If i'm player 0, then search over all possible moves, assume player 2
    takes the agree action, and choose the action that maximizes the NBS
    Player 1 just always agrees.

    Args:
      state: the OpenSpiel state to act from.
    """
    player = state.current_player()
    if player == 1:
      return state.agree_action()
    max_nbs = -1
    max_action = -1
    for action in state.legal_actions():
      state_clone = state.clone()
      state_clone.apply_action(action)
      state_clone.apply_action(state.agree_action())
      returns = state_clone.returns()
      nbs = np.prod(returns)
      if nbs > max_nbs:
        max_nbs = nbs
        max_action = action
    assert max_action >= 0
    return max_action


def main(_):
  assert FLAGS.data_file is not None
  _, negotiations = parse_dataset(FLAGS.data_file)

  print(f"Writing instances database: {FLAGS.instances_file}")
  write_instances_file(negotiations, FLAGS.instances_file)

  # Human NBS
  avg_human_nbs = 0
  for neg in negotiations:
    avg_human_nbs += np.prod(neg.rewards)
  avg_human_nbs /= len(negotiations)
  print(f"Average human NBS: {avg_human_nbs}")

  game = pyspiel.load_game("bargaining",
                           {"instances_file": FLAGS.instances_file})

  # Max bot
  bots = [MaxBot(), MaxBot()]
  avg_max_nbs = compute_nbs_from_simulations(game, 6796, bots)
  print(f"Average max NBS: {avg_max_nbs}")

  # Uniform random NBS
  bots = [
      pyspiel.make_uniform_random_bot(0, np.random.randint(0, 1000000)),
      pyspiel.make_uniform_random_bot(1, np.random.randint(0, 1000000)),
  ]
  avg_uniform_nbs = compute_nbs_from_simulations(game, 6796, bots)
  print(f"Average uniform NBS: {avg_uniform_nbs}")

  # IS-MCTS NBS
  evaluator = pyspiel.RandomRolloutEvaluator(1, np.random.randint(0, 1000000))
  bots = [
      pyspiel.ISMCTSBot(
          np.random.randint(0, 1000000), evaluator, 10.0, 1000, -1,
          pyspiel.ISMCTSFinalPolicyType.MAX_VISIT_COUNT, False, False),
      pyspiel.ISMCTSBot(
          np.random.randint(0, 1000000), evaluator, 10.0, 1000, -1,
          pyspiel.ISMCTSFinalPolicyType.MAX_VISIT_COUNT, False, False)
  ]
  avg_ismcts_nbs = compute_nbs_from_simulations(game, 6796, bots)
  print(f"Average IS-MCTS NBS: {avg_ismcts_nbs}")


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tabular Multiagent Q-learning agent.

Currently implementations include:
Nash-Q: https://www.jmlr.org/papers/volume4/hu03a/hu03a.pdf
Correlated-Q: https://www.aaai.org/Papers/ICML/2003/ICML03-034.pdf, where both
CE-Q and CCE-Q are supported.
"""

import abc
import collections
import itertools
import numpy as np

from open_spiel.python import rl_agent
from open_spiel.python import rl_tools
from open_spiel.python.algorithms.jpsro import _mgcce
from open_spiel.python.algorithms.jpsro import _mgce
from open_spiel.python.algorithms.matrix_nash import lemke_howson_solve


def valuedict():
  return collections.defaultdict(float)


class JointActionSolver:

  @abc.abstractmethod
  def __call__(self, payoffs_array):
    """Find a joint action mixture and values for the current one-step game.

    Args:
      payoffs_array: a `numpy.ndarray` of utilities of a game.

    Returns:
      res_mixtures: a list of mixed strategies for each agent
      res_values: a list of expected utilities for each agent
    """


class TwoPlayerNashSolver(JointActionSolver):
  """A joint action solver solving for Nash for two-player games.

  Uses python.algorithms.matrix_nash.lemke_howson_solve
  """

  def __call__(self, payoffs_array):
    assert len(payoffs_array) == 2

    row_payoffs, col_payoffs = payoffs_array[0], payoffs_array[1]
    a0, a1 = payoffs_array.shape[1:]

    best_value = float("-inf")
    res_mixtures, res_values = None, None

    for (row_mixture,
         col_mixture) in lemke_howson_solve(row_payoffs, col_payoffs):
      # TO-DO: handle the case where the LH solver gave ineligible answer
      if np.sum(np.isnan(row_mixture)) or np.sum(np.isnan(col_mixture)):
        continue
      row_mixture_, col_mixture_ = row_mixture.reshape(
          (-1, 1)), col_mixture.reshape((-1, 1))
      row_value, col_value = (
          row_mixture_.T.dot(row_payoffs).dot(col_mixture_)).item(), (
              row_mixture_.T.dot(col_payoffs).dot(col_mixture_)).item()
      # Currently using maximizing social welfare for equilibrium selection
      if row_value + col_value > best_value:
        best_value = row_value + col_value
        res_mixtures = [row_mixture, col_mixture]
        res_values = [row_value, col_value]

    # If no plauisble nash found, use uniform mixed strategies
    if not res_mixtures:
      res_mixtures = [np.ones(a0) / a0, np.ones(a1) / a1]
      row_mixture_, col_mixture_ = res_mixtures[0].reshape(
          (-1, 1)), res_mixtures[1].reshape((-1, 1))
      res_values = [(row_mixture_.T.dot(row_payoffs).dot(col_mixture_)).item(),
                    (row_mixture_.T.dot(col_payoffs).dot(col_mixture_)).item()]

    return res_mixtures, res_values


class CorrelatedEqSolver(JointActionSolver):
  """A joint action solver solving for correlated equilibrium.

  Uses python.algorithms.jspro._mgce and _mgcce for solving (coarse) correlated
  equilibrium.
  """

  def __init__(self, is_cce=False):
    self._is_cce = is_cce

  def __call__(self, payoffs_array):
    size = len(payoffs_array)
    assert size > 0
    mixture, _ = (_mgcce(payoffs_array, [1] * size, ignore_repeats=True) if
                  self._is_cce else _mgce(payoffs_array, [1] * size,
                                          ignore_repeats=True))
    mixtures, values = [], []
    for n in range(size):
      values.append(np.sum(payoffs_array[n] * mixture))
      mixtures.append(
          np.sum(mixture, axis=tuple([n_ for n_ in range(size) if n_ != n])))
    return mixtures, values


class MultiagentQLearner(rl_agent.AbstractAgent):
  """A multiagent joint action learner."""

  def __init__(self,
               player_id,
               num_players,
               num_actions,
               joint_action_solver,
               step_size=0.1,
               epsilon_schedule=rl_tools.ConstantSchedule(0.2),
               discount_factor=1.0):
    """Initialize the Multiagent joint-action Q-Learning agent.

    The joint_action_solver solves for one-step matrix game defined by Q-tables.

    Args:
      player_id: the player id this agent will play as,
      num_players: the number of players in the game,
      num_actions: the number of distinct actions in the game,
      joint_action_solver: the joint action solver class to use to solve the
        one-step matrix games
      step_size: learning rate for Q-learning,
      epsilon_schedule: exploration parameter,
      discount_factor: the discount factor as in Q-learning.
    """
    self._player_id = player_id
    self._num_players = num_players
    self._num_actions = num_actions
    self._joint_action_solver = joint_action_solver
    self._step_size = step_size
    self._epsilon_schedule = epsilon_schedule
    self._epsilon = epsilon_schedule.value
    self._discount_factor = discount_factor
    self._q_values = [
        collections.defaultdict(valuedict) for _ in range(num_players)
    ]
    self._prev_info_state = None

  def restart_episode(self):
    self._prev_info_state = None

  def _get_payoffs_array(self, info_state):
    payoffs_array = np.zeros((self._num_players,) + tuple(self._num_actions))
    for joint_action in itertools.product(
        *[range(dim) for dim in self._num_actions]):
      for n in range(self._num_players):
        payoffs_array[
            (n,) + joint_action] = self._q_values[n][info_state][joint_action]
    return payoffs_array

  def _epsilon_greedy(self, info_state, legal_actions, epsilon):
    """Returns a valid epsilon-greedy action and valid action probs.

    If the agent has not been to `info_state`, a valid random action is chosen.
    Args:
      info_state: hashable representation of the information state.
      legal_actions: list of actions at `info_state`.
      epsilon: float, prob of taking an exploratory action.

    Returns:
      A valid epsilon-greedy action and valid action probabilities.
    """
    probs = np.zeros(self._num_actions[self._player_id])

    state_probs, _ = self._joint_action_solver(
        self._get_payoffs_array(info_state))

    probs[legal_actions[self._player_id]] = (
        epsilon / len(legal_actions[self._player_id]))
    probs += (1 - epsilon) * state_probs[self._player_id]
    action = np.random.choice(
        range(self._num_actions[self._player_id]), p=probs)
    return action, probs

  def step(self, time_step, actions=None, is_evaluation=False):
    """Returns the action to be taken and updates the Q-values if needed.

    Args:
        time_step: an instance of rl_environment.TimeStep,
        actions: list of actions taken by all agents from the previous step,
        is_evaluation: bool, whether this is a training or evaluation call,

    Returns:
        A `rl_agent.StepOutput` containing the action probs and chosen action.
    """
    info_state = str(time_step.observations["info_state"])
    legal_actions = time_step.observations["legal_actions"]

    # Prevent undefined errors if this agent never plays until terminal step
    action, probs = None, None

    # Act step: don't act at terminal states.
    if not time_step.last():
      epsilon = 0.0 if is_evaluation else self._epsilon
      # select according to the joint action solver
      action, probs = self._epsilon_greedy(
          info_state, legal_actions, epsilon=epsilon)

    # Learn step: don't learn during evaluation or at first agent steps.
    actions = tuple(actions)

    if self._prev_info_state and not is_evaluation:
      _, next_state_values = (
          self._joint_action_solver(self._get_payoffs_array(info_state)))
      # update Q values for every agent
      for n in range(self._num_players):
        target = time_step.rewards[n]
        if not time_step.last():  # Q values are zero for terminal.
          target += self._discount_factor * next_state_values[n]

        prev_q_value = self._q_values[n][self._prev_info_state][actions]

        self._q_values[n][self._prev_info_state][actions] += (
            self._step_size * (target - prev_q_value))

      # Decay epsilon, if necessary.
      self._epsilon = self._epsilon_schedule.step()

      if time_step.last():  # prepare for the next episode.
        self._prev_info_state = None
        return

    # Don't mess up with the state during evaluation.
    if not is_evaluation:
      self._prev_info_state = info_state

    return rl_agent.StepOutput(action=action, probs=probs)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tabular Q-Learner example on Tic Tac Toe.

Two Q-Learning agents are trained by playing against each other. Then, the game
can be played against the agents from the command line.

After about 10**5 training episodes, the agents reach a good policy: win rate
against random opponents is around 99% for player 0 and 92% for player 1.
"""

import logging
import sys
from absl import app
from absl import flags
import numpy as np

from open_spiel.python import rl_environment
from open_spiel.python.algorithms import random_agent
from open_spiel.python.algorithms import tabular_qlearner

FLAGS = flags.FLAGS

flags.DEFINE_integer("num_episodes", int(5e4), "Number of train episodes.")
flags.DEFINE_boolean(
    "iteractive_play", True,
    "Whether to run an interactive play with the agent after training.")


def pretty_board(time_step):
  """Returns the board in `time_step` in a human readable format."""
  info_state = time_step.observations["info_state"][0]
  x_locations = np.nonzero(info_state[9:18])[0]
  o_locations = np.nonzero(info_state[18:])[0]
  board = np.full(3 * 3, ".")
  board[x_locations] = "X"
  board[o_locations] = "0"
  board = np.reshape(board, (3, 3))
  return board


def command_line_action(time_step):
  """Gets a valid action from the user on the command line."""
  current_player = time_step.observations["current_player"]
  legal_actions = time_step.observations["legal_actions"][current_player]
  action = -1
  while action not in legal_actions:
    print("Choose an action from {}:".format(legal_actions))
    sys.stdout.flush()
    action_str = input()
    try:
      action = int(action_str)
    except ValueError:
      continue
  return action


def eval_against_random_bots(env, trained_agents, random_agents, num_episodes):
  """Evaluates `trained_agents` against `random_agents` for `num_episodes`."""
  wins = np.zeros(2)
  for player_pos in range(2):
    if player_pos == 0:
      cur_agents = [trained_agents[0], random_agents[1]]
    else:
      cur_agents = [random_agents[0], trained_agents[1]]
    for _ in range(num_episodes):
      time_step = env.reset()
      while not time_step.last():
        player_id = time_step.observations["current_player"]
        agent_output = cur_agents[player_id].step(time_step, is_evaluation=True)
        time_step = env.step([agent_output.action])
      if time_step.rewards[player_pos] > 0:
        wins[player_pos] += 1
  return wins / num_episodes


def main(_):
  game = "tic_tac_toe"
  num_players = 2

  env = rl_environment.Environment(game)
  num_actions = env.action_spec()["num_actions"]

  agents = [
      tabular_qlearner.QLearner(player_id=idx, num_actions=num_actions)
      for idx in range(num_players)
  ]

  # random agents for evaluation
  random_agents = [
      random_agent.RandomAgent(player_id=idx, num_actions=num_actions)
      for idx in range(num_players)
  ]

  # 1. Train the agents
  training_episodes = FLAGS.num_episodes
  for cur_episode in range(training_episodes):
    if cur_episode % int(1e4) == 0:
      win_rates = eval_against_random_bots(env, agents, random_agents, 1000)
      logging.info("Starting episode %s, win_rates %s", cur_episode, win_rates)
    time_step = env.reset()
    while not time_step.last():
      player_id = time_step.observations["current_player"]
      agent_output = agents[player_id].step(time_step)
      time_step = env.step([agent_output.action])

    # Episode is over, step all agents with final info state.
    for agent in agents:
      agent.step(time_step)

  if not FLAGS.iteractive_play:
    return

  # 2. Play from the command line against the trained agent.
  human_player = 1
  while True:
    logging.info("You are playing as %s", "O" if human_player else "X")
    time_step = env.reset()
    while not time_step.last():
      player_id = time_step.observations["current_player"]
      if player_id == human_player:
        agent_out = agents[human_player].step(time_step, is_evaluation=True)
        logging.info("\n%s", agent_out.probs.reshape((3, 3)))
        logging.info("\n%s", pretty_board(time_step))
        action = command_line_action(time_step)
      else:
        agent_out = agents[1 - human_player].step(time_step, is_evaluation=True)
        action = agent_out.action
      time_step = env.step([action])

    logging.info("\n%s", pretty_board(time_step))

    logging.info("End of game!")
    if time_step.rewards[human_player] > 0:
      logging.info("You win")
    elif time_step.rewards[human_player] < 0:
      logging.info("You lose")
    else:
      logging.info("Draw")
    # Switch order of players
    human_player = 1 - human_player


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A simple random bot."""

import base64
import sys
import numpy as np
from open_spiel.python.observation import make_observation
import pyspiel

# Example implementation of the random bot for the HIG competition.
# The bot must strictly follow the communication protocol via stdin/stdout,
# but it can print any message to stderr for debugging.

# Read the current setup.
game_name = input()
play_as = int(input())

print(game_name, play_as, file=sys.stderr)  # For debugging purposes.

# Load the provided game.
game = pyspiel.load_game(game_name)

# Observations will be received later from the referee.
# The referee factors the observation into public (common knowledge across all
# players) and private parts.
public_observation = make_observation(
    game,
    pyspiel.IIGObservationType(
        perfect_recall=False,
        public_info=True,
        private_info=pyspiel.PrivateInfoType.NONE))
private_observation = make_observation(
    game,
    pyspiel.IIGObservationType(
        perfect_recall=False,
        public_info=False,
        private_info=pyspiel.PrivateInfoType.SINGLE_PLAYER))

# Now there is 5 secs warm-up time that could be used for loading relevant
# supplementary data. All data can be read/written from persistent /data
# directory mounted from an external storage.
print("ready")

# Loop per match. This loop will end when referee instructs the player to do so.
while True:

  # Acknowledge the match started.
  print("start")

  # This is just a placeholder for other implementations -- we do not use
  # state in random agent, as it receives list of actions it can pick from.
  state = game.new_initial_state()

  while True:  # Loop per state in match.
    message = input()  # Read message from the referee.
    print(message, file=sys.stderr)  # For debugging purposes.

    if message == "tournament over":
      # The tournament is now over: there is 60 sec shutdown time
      # available for processing tournament results by the agent,
      # for example to update supplementary data.
      print("tournament over")
      sys.exit(0)

    if message.startswith("match over"):
      # The full message has format "game over 123"
      # where 123 is the final float reward received by this bot.
      #
      # Note that this message does not necessarily mean the match
      # reached a terminal state: if opponent crashed / violated
      # rules, the match will be over as well.
      print("match over")
      break

    # Regular message: a public and private observation followed by
    # a list of legal actions (if the bot should be acting).
    public_buf, private_buf, *legal_actions = message.split(" ")
    public_observation.decompress(base64.b64decode(public_buf))
    private_observation.decompress(base64.b64decode(private_buf))

    if legal_actions:
      # There is time limit of 5 secs.
      print(np.random.choice(legal_actions))
    else:
      # Pondering phase, i.e. thinking when the bot is not acting.
      # The time limit is always at least 0.2s, but can be longer,
      # up to 5s, depending on how long the opponent thinks.
      print("ponder")  # This bot does not ponder.

  assert message.startswith("match over")
  score = int(message.split(" ")[-1])
  print("score:", score, file=sys.stderr)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Game-specific query example."""

from absl import app
from absl import flags

import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_string("game", "negotiation", "Name of the game")


def main(_):
  print("Creating game: " + FLAGS.game)
  game = pyspiel.load_game(FLAGS.game)

  state = game.new_initial_state()

  print(str(state))

  # Need to apply the first chance node for items and utilities to be generated
  state.apply_action(0)

  print("Item pool: {}".format(state.item_pool()))
  print("Player 0 utils: {}".format(state.agent_utils(0)))
  print("Player 1 utils: {}".format(state.agent_utils(1)))

  state = game.new_initial_state()

  print(str(state))

  # Need to apply the first chance node for items and utilities to be generated
  state.apply_action(0)

  print("Item pool: {}".format(state.item_pool()))
  print("Player 0 utils: {}".format(state.agent_utils(0)))
  print("Player 1 utils: {}".format(state.agent_utils(1)))


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Exploitability measurement utils."""

from absl import logging  # pylint:disable=unused-import

import numpy as np
from scipy import special

from open_spiel.python.algorithms.adidas_utils.helpers import simplex
from open_spiel.python.algorithms.adidas_utils.helpers.symmetric import exploitability


class Solver(object):
  """Generic Solver."""

  def __init__(self, proj_grad=True, euclidean=False, rnd_init=False,
               seed=None):
    """Ctor."""
    self.num_players = None
    self.proj_grad = proj_grad
    self.rnd_init = rnd_init
    self.lrs = (None, None, None)
    self.has_aux = False

    self.euclidean = euclidean
    if euclidean:
      self.update = self.euc_descent_step
    else:
      self.update = self.mirror_descent_step

    self.seed = seed
    self.random = np.random.RandomState(seed)

  def init_vars(self, num_strats, num_players):
    """Initialize solver parameters."""
    self.num_players = num_players
    if self.rnd_init:
      init_dist = self.random.rand(num_strats)
    else:
      init_dist = np.ones(num_strats)
    init_dist /= init_dist.sum()
    return (init_dist,)

  def compute_gradients(self, params, payoff_matrices):
    """Compute and return gradients for all parameters.

    Args:
      params: e.g., tuple of params (dist,)
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      eg., tuple of gradients (grad_dist,)
    """
    raise NotImplementedError("Should be implemented by specific solver.")

  def exploitability(self, params, payoff_matrices):
    """Compute and return exploitability that solver is minimizing.

    Args:
      params: e.g., tuple of params (dist,)
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      float, exploitability of current dist
    """
    return exploitability.unreg_exploitability(params, payoff_matrices)

  def euc_descent_step(self, params, grads, t):
    """Projected gradient descent on exploitability using Euclidean projection.

    Args:
      params: tuple of variables to be updated (dist,)
      grads: tuple of variable gradients (grad_dist,)
      t: int, solver iteration
    Returns:
      new_params: tuple of update params (new_dist,)
    """
    del t
    new_params = params[0] - self.lrs[0] * grads[0]
    new_params = simplex.euclidean_projection_onto_simplex(new_params)
    return (new_params,)

  def mirror_descent_step(self, params, grads, t):
    """Entropic mirror descent on exploitability.

    Args:
      params: tuple of variables to be updated (dist)
      grads: tuple of variable gradients (grad_dist)
      t: int, solver iteration
    Returns:
      new_params: tuple of update params (new_dist)
    """
    del t
    dist = np.clip(params[0], 0, np.inf)
    return (special.softmax(np.log(dist) - self.lrs[0] * grads[0]),)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An AlphaZero style model with a policy and value head."""

import collections
import functools
import os
from typing import Sequence

import numpy as np
import tensorflow.compat.v1 as tf


def cascade(x, fns):
  for fn in fns:
    x = fn(x)
  return x

tfkl = tf.keras.layers
conv_2d = functools.partial(tfkl.Conv2D, padding="same")


def batch_norm(training, updates, name):
  """A batch norm layer.

  Args:
    training: A placeholder of whether this is done in training or not.
    updates: A list to be extended with this layer's updates.
    name: Name of the layer.

  Returns:
    A function to apply to the previous layer.
  """
  bn = tfkl.BatchNormalization(name=name)
  def batch_norm_layer(x):
    # This emits a warning that training is a placeholder instead of a concrete
    # bool, but seems to work anyway.
    applied = bn(x, training)
    updates.extend(bn.updates)
    return applied
  return batch_norm_layer


def residual_layer(inputs, num_filters, kernel_size, training, updates, name):
  return cascade(inputs, [
      conv_2d(num_filters, kernel_size, name=f"{name}_res_conv1"),
      batch_norm(training, updates, f"{name}_res_batch_norm1"),
      tfkl.Activation("relu"),
      conv_2d(num_filters, kernel_size, name=f"{name}_res_conv2"),
      batch_norm(training, updates, f"{name}_res_batch_norm2"),
      lambda x: tfkl.add([x, inputs]),
      tfkl.Activation("relu"),
  ])


class TrainInput(collections.namedtuple(
    "TrainInput", "observation legals_mask policy value")):
  """Inputs for training the Model."""

  @staticmethod
  def stack(train_inputs):
    observation, legals_mask, policy, value = zip(*train_inputs)
    return TrainInput(
        np.array(observation, dtype=np.float32),
        np.array(legals_mask, dtype=np.bool),
        np.array(policy),
        np.expand_dims(value, 1))


class Losses(collections.namedtuple("Losses", "policy value l2")):
  """Losses from a training step."""

  @property
  def total(self):
    return self.policy + self.value + self.l2

  def __str__(self):
    return ("Losses(total: {:.3f}, policy: {:.3f}, value: {:.3f}, "
            "l2: {:.3f})").format(self.total, self.policy, self.value, self.l2)

  def __add__(self, other):
    return Losses(self.policy + other.policy,
                  self.value + other.value,
                  self.l2 + other.l2)

  def __truediv__(self, n):
    return Losses(self.policy / n, self.value / n, self.l2 / n)


class Model(object):
  """An AlphaZero style model with a policy and value head.

  This supports three types of models: mlp, conv2d and resnet.

  All models have a shared torso stack with two output heads: policy and value.
  They have same meaning as in the AlphaGo Zero and AlphaZero papers. The resnet
  model copies the one in that paper when set with width 256 and depth 20. The
  conv2d model is the same as the resnet except uses a conv+batchnorm+relu
  instead of the res blocks. The mlp model uses dense layers instead of conv,
  and drops batch norm.

  Links to relevant articles/papers:
    https://deepmind.com/blog/article/alphago-zero-starting-scratch has an open
      access link to the AlphaGo Zero nature paper.
    https://deepmind.com/blog/article/alphazero-shedding-new-light-grand-games-chess-shogi-and-go
      has an open access link to the AlphaZero science paper.

  All are parameterized by their input (observation) shape and output size
  (number of actions), though the conv2d and resnet might only work with games
  that have spatial data (ie 3 non-batch dimensions, eg: connect four would
  work, but not poker).

  The depth is the number of blocks in the torso, where the definition of a
  block varies by model. For a resnet it's a resblock which is two conv2ds,
  batch norms and relus, and an addition. For conv2d it's a conv2d, a batch norm
  and a relu. For mlp it's a dense plus relu.

  The width is the number of filters for any conv2d and the number of hidden
  units for any dense layer.

  Note that this uses an explicit graph so that it can be used for inference
  and training from C++. It seems to also be 20%+ faster than using eager mode,
  at least for the unit test.
  """

  valid_model_types = ["mlp", "conv2d", "resnet"]

  def __init__(self, session, saver, path):
    """Init a model. Use build_model, from_checkpoint or from_graph instead."""
    self._session = session
    self._saver = saver
    self._path = path

    def get_var(name):
      return self._session.graph.get_tensor_by_name(name + ":0")

    self._input = get_var("input")
    self._legals_mask = get_var("legals_mask")
    self._training = get_var("training")
    self._value_out = get_var("value_out")
    self._policy_softmax = get_var("policy_softmax")
    self._policy_loss = get_var("policy_loss")
    self._value_loss = get_var("value_loss")
    self._l2_reg_loss = get_var("l2_reg_loss")
    self._policy_targets = get_var("policy_targets")
    self._value_targets = get_var("value_targets")
    self._train = self._session.graph.get_operation_by_name("train")

  @classmethod
  def build_model(cls, model_type, input_shape, output_size, nn_width, nn_depth,
                  weight_decay, learning_rate, path):
    """Build a model with the specified params."""
    if model_type not in cls.valid_model_types:
      raise ValueError(f"Invalid model type: {model_type}, "
                       f"expected one of: {cls.valid_model_types}")

    # The order of creating the graph, init, saver, and session is important!
    # https://stackoverflow.com/a/40788998
    g = tf.Graph()  # Allow multiple independent models and graphs.
    with g.as_default():
      cls._define_graph(model_type, input_shape, output_size, nn_width,
                        nn_depth, weight_decay, learning_rate)
      init = tf.variables_initializer(tf.global_variables(),
                                      name="init_all_vars_op")
      with tf.device("/cpu:0"):  # Saver only works on CPU.
        saver = tf.train.Saver(
            max_to_keep=10000, sharded=False, name="saver")
    session = tf.Session(graph=g)
    session.__enter__()
    session.run(init)
    return cls(session, saver, path)

  @classmethod
  def from_checkpoint(cls, checkpoint, path=None):
    """Load a model from a checkpoint."""
    model = cls.from_graph(checkpoint, path)
    model.load_checkpoint(checkpoint)
    return model

  @classmethod
  def from_graph(cls, metagraph, path=None):
    """Load only the model from a graph or checkpoint."""
    if not os.path.exists(metagraph):
      metagraph += ".meta"
    if not path:
      path = os.path.dirname(metagraph)
    g = tf.Graph()  # Allow multiple independent models and graphs.
    with g.as_default():
      saver = tf.train.import_meta_graph(metagraph)
    session = tf.Session(graph=g)
    session.__enter__()
    session.run("init_all_vars_op")
    return cls(session, saver, path)

  def __del__(self):
    if hasattr(self, "_session") and self._session:
      self._session.close()

  @staticmethod
  def _define_graph(model_type, input_shape, output_size,
                    nn_width, nn_depth, weight_decay, learning_rate):
    """Define the model graph."""
    # Inference inputs
    input_size = int(np.prod(input_shape))
    observations = tf.placeholder(tf.float32, [None, input_size], name="input")
    legals_mask = tf.placeholder(tf.bool, [None, output_size],
                                 name="legals_mask")
    training = tf.placeholder(tf.bool, name="training")

    bn_updates = []

    # Main torso of the network
    if model_type == "mlp":
      torso = observations  # Ignore the input shape, treat it as a flat array.
      for i in range(nn_depth):
        torso = cascade(torso, [
            tfkl.Dense(nn_width, name=f"torso_{i}_dense"),
            tfkl.Activation("relu"),
        ])
    elif model_type == "conv2d":
      torso = tfkl.Reshape(input_shape)(observations)
      for i in range(nn_depth):
        torso = cascade(torso, [
            conv_2d(nn_width, 3, name=f"torso_{i}_conv"),
            batch_norm(training, bn_updates, f"torso_{i}_batch_norm"),
            tfkl.Activation("relu"),
        ])
    elif model_type == "resnet":
      torso = cascade(observations, [
          tfkl.Reshape(input_shape),
          conv_2d(nn_width, 3, name="torso_in_conv"),
          batch_norm(training, bn_updates, "torso_in_batch_norm"),
          tfkl.Activation("relu"),
      ])
      for i in range(nn_depth):
        torso = residual_layer(torso, nn_width, 3, training, bn_updates,
                               f"torso_{i}")
    else:
      raise ValueError("Unknown model type.")

    # The policy head
    if model_type == "mlp":
      policy_head = cascade(torso, [
          tfkl.Dense(nn_width, name="policy_dense"),
          tfkl.Activation("relu"),
      ])
    else:
      policy_head = cascade(torso, [
          conv_2d(filters=2, kernel_size=1, name="policy_conv"),
          batch_norm(training, bn_updates, "policy_batch_norm"),
          tfkl.Activation("relu"),
          tfkl.Flatten(),
      ])
    policy_logits = tfkl.Dense(output_size, name="policy")(policy_head)
    policy_logits = tf.where(legals_mask, policy_logits,
                             -1e32 * tf.ones_like(policy_logits))
    unused_policy_softmax = tf.identity(tfkl.Softmax()(policy_logits),
                                        name="policy_softmax")
    policy_targets = tf.placeholder(
        shape=[None, output_size], dtype=tf.float32, name="policy_targets")
    policy_loss = tf.reduce_mean(
        tf.nn.softmax_cross_entropy_with_logits_v2(
            logits=policy_logits, labels=policy_targets),
        name="policy_loss")

    # The value head
    if model_type == "mlp":
      value_head = torso  # Nothing specific before the shared value head.
    else:
      value_head = cascade(torso, [
          conv_2d(filters=1, kernel_size=1, name="value_conv"),
          batch_norm(training, bn_updates, "value_batch_norm"),
          tfkl.Activation("relu"),
          tfkl.Flatten(),
      ])
    value_out = cascade(value_head, [
        tfkl.Dense(nn_width, name="value_dense"),
        tfkl.Activation("relu"),
        tfkl.Dense(1, name="value"),
        tfkl.Activation("tanh"),
    ])
    # Need the identity to name the single value output from the dense layer.
    value_out = tf.identity(value_out, name="value_out")
    value_targets = tf.placeholder(
        shape=[None, 1], dtype=tf.float32, name="value_targets")
    value_loss = tf.identity(tf.losses.mean_squared_error(
        value_out, value_targets), name="value_loss")

    l2_reg_loss = tf.add_n([
        weight_decay * tf.nn.l2_loss(var)
        for var in tf.trainable_variables()
        if "/bias:" not in var.name
    ], name="l2_reg_loss")

    total_loss = policy_loss + value_loss + l2_reg_loss
    optimizer = tf.train.AdamOptimizer(learning_rate)
    with tf.control_dependencies(bn_updates):
      unused_train = optimizer.minimize(total_loss, name="train")

  @property
  def num_trainable_variables(self):
    return sum(np.prod(v.shape) for v in tf.trainable_variables())

  def print_trainable_variables(self):
    for v in tf.trainable_variables():
      print("{}: {}".format(v.name, v.shape))

  def write_graph(self, filename):
    full_path = os.path.join(self._path, filename)
    tf.train.export_meta_graph(
        graph_def=self._session.graph_def, saver_def=self._saver.saver_def,
        filename=full_path, as_text=False)
    return full_path

  def inference(self, observation, legals_mask):
    return self._session.run(
        [self._value_out, self._policy_softmax],
        feed_dict={self._input: np.array(observation, dtype=np.float32),
                   self._legals_mask: np.array(legals_mask, dtype=np.bool),
                   self._training: False})

  def update(self, train_inputs: Sequence[TrainInput]):
    """Runs a training step."""
    batch = TrainInput.stack(train_inputs)

    # Run a training step and get the losses.
    _, policy_loss, value_loss, l2_reg_loss = self._session.run(
        [self._train, self._policy_loss, self._value_loss, self._l2_reg_loss],
        feed_dict={self._input: batch.observation,
                   self._legals_mask: batch.legals_mask,
                   self._policy_targets: batch.policy,
                   self._value_targets: batch.value,
                   self._training: True})

    return Losses(policy_loss, value_loss, l2_reg_loss)

  def save_checkpoint(self, step):
    return self._saver.save(
        self._session,
        os.path.join(self._path, "checkpoint"),
        global_step=step)

  def load_checkpoint(self, path):
    return self._saver.restore(self._session, path)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MCTS example."""

import collections
import random
import sys

from absl import app
from absl import flags
import numpy as np

from open_spiel.python.algorithms import mcts
from open_spiel.python.algorithms.alpha_zero import evaluator as az_evaluator
from open_spiel.python.algorithms.alpha_zero import model as az_model
from open_spiel.python.bots import gtp
from open_spiel.python.bots import human
from open_spiel.python.bots import uniform_random
import pyspiel

_KNOWN_PLAYERS = [
    # A generic Monte Carlo Tree Search agent.
    "mcts",

    # A generic random agent.
    "random",

    # You'll be asked to provide the moves.
    "human",

    # Run an external program that speaks the Go Text Protocol.
    # Requires the gtp_path flag.
    "gtp",

    # Run an alpha_zero checkpoint with MCTS. Uses the specified UCT/sims.
    # Requires the az_path flag.
    "az"
]

flags.DEFINE_string("game", "tic_tac_toe", "Name of the game.")
flags.DEFINE_enum("player1", "mcts", _KNOWN_PLAYERS, "Who controls player 1.")
flags.DEFINE_enum("player2", "random", _KNOWN_PLAYERS, "Who controls player 2.")
flags.DEFINE_string("gtp_path", None, "Where to find a binary for gtp.")
flags.DEFINE_multi_string("gtp_cmd", [], "GTP commands to run at init.")
flags.DEFINE_string("az_path", None,
                    "Path to an alpha_zero checkpoint. Needed by an az player.")
flags.DEFINE_integer("uct_c", 2, "UCT's exploration constant.")
flags.DEFINE_integer("rollout_count", 1, "How many rollouts to do.")
flags.DEFINE_integer("max_simulations", 1000, "How many simulations to run.")
flags.DEFINE_integer("num_games", 1, "How many games to play.")
flags.DEFINE_integer("seed", None, "Seed for the random number generator.")
flags.DEFINE_bool("random_first", False, "Play the first move randomly.")
flags.DEFINE_bool("solve", True, "Whether to use MCTS-Solver.")
flags.DEFINE_bool("quiet", False, "Don't show the moves as they're played.")
flags.DEFINE_bool("verbose", False, "Show the MCTS stats of possible moves.")

FLAGS = flags.FLAGS


def _opt_print(*args, **kwargs):
  if not FLAGS.quiet:
    print(*args, **kwargs)


def _init_bot(bot_type, game, player_id):
  """Initializes a bot by type."""
  rng = np.random.RandomState(FLAGS.seed)
  if bot_type == "mcts":
    evaluator = mcts.RandomRolloutEvaluator(FLAGS.rollout_count, rng)
    return mcts.MCTSBot(
        game,
        FLAGS.uct_c,
        FLAGS.max_simulations,
        evaluator,
        random_state=rng,
        solve=FLAGS.solve,
        verbose=FLAGS.verbose)
  if bot_type == "az":
    model = az_model.Model.from_checkpoint(FLAGS.az_path)
    evaluator = az_evaluator.AlphaZeroEvaluator(game, model)
    return mcts.MCTSBot(
        game,
        FLAGS.uct_c,
        FLAGS.max_simulations,
        evaluator,
        random_state=rng,
        child_selection_fn=mcts.SearchNode.puct_value,
        solve=FLAGS.solve,
        verbose=FLAGS.verbose)
  if bot_type == "random":
    return uniform_random.UniformRandomBot(player_id, rng)
  if bot_type == "human":
    return human.HumanBot()
  if bot_type == "gtp":
    bot = gtp.GTPBot(game, FLAGS.gtp_path)
    for cmd in FLAGS.gtp_cmd:
      bot.gtp_cmd(cmd)
    return bot
  raise ValueError("Invalid bot type: %s" % bot_type)


def _get_action(state, action_str):
  for action in state.legal_actions():
    if action_str == state.action_to_string(state.current_player(), action):
      return action
  return None


def _play_game(game, bots, initial_actions):
  """Plays one game."""
  state = game.new_initial_state()
  _opt_print("Initial state:\n{}".format(state))

  history = []

  if FLAGS.random_first:
    assert not initial_actions
    initial_actions = [state.action_to_string(
        state.current_player(), random.choice(state.legal_actions()))]

  for action_str in initial_actions:
    action = _get_action(state, action_str)
    if action is None:
      sys.exit("Invalid action: {}".format(action_str))

    history.append(action_str)
    for bot in bots:
      bot.inform_action(state, state.current_player(), action)
    state.apply_action(action)
    _opt_print("Forced action", action_str)
    _opt_print("Next state:\n{}".format(state))

  while not state.is_terminal():
    current_player = state.current_player()
    # The state can be three different types: chance node,
    # simultaneous node, or decision node
    if state.is_chance_node():
      # Chance node: sample an outcome
      outcomes = state.chance_outcomes()
      num_actions = len(outcomes)
      _opt_print("Chance node, got " + str(num_actions) + " outcomes")
      action_list, prob_list = zip(*outcomes)
      action = np.random.choice(action_list, p=prob_list)
      action_str = state.action_to_string(current_player, action)
      _opt_print("Sampled action: ", action_str)
    elif state.is_simultaneous_node():
      raise ValueError("Game cannot have simultaneous nodes.")
    else:
      # Decision node: sample action for the single current player
      bot = bots[current_player]
      action = bot.step(state)
      action_str = state.action_to_string(current_player, action)
      _opt_print("Player {} sampled action: {}".format(current_player,
                                                       action_str))

    for i, bot in enumerate(bots):
      if i != current_player:
        bot.inform_action(state, current_player, action)
    history.append(action_str)
    state.apply_action(action)

    _opt_print("Next state:\n{}".format(state))

  # Game is now done. Print return for each player
  returns = state.returns()
  print("Returns:", " ".join(map(str, returns)), ", Game actions:",
        " ".join(history))

  for bot in bots:
    bot.restart()

  return returns, history


def main(argv):
  game = pyspiel.load_game(FLAGS.game)
  if game.num_players() > 2:
    sys.exit("This game requires more players than the example can handle.")
  bots = [
      _init_bot(FLAGS.player1, game, 0),
      _init_bot(FLAGS.player2, game, 1),
  ]
  histories = collections.defaultdict(int)
  overall_returns = [0, 0]
  overall_wins = [0, 0]
  game_num = 0
  try:
    for game_num in range(FLAGS.num_games):
      returns, history = _play_game(game, bots, argv[1:])
      histories[" ".join(history)] += 1
      for i, v in enumerate(returns):
        overall_returns[i] += v
        if v > 0:
          overall_wins[i] += 1
  except (KeyboardInterrupt, EOFError):
    game_num -= 1
    print("Caught a KeyboardInterrupt, stopping early.")
  print("Number of games played:", game_num + 1)
  print("Number of distinct games played:", len(histories))
  print("Players:", FLAGS.player1, FLAGS.player2)
  print("Overall wins", overall_wins)
  print("Overall returns", overall_returns)


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Compute the exploitability of a bot / strategy in a 2p sequential game.

This computes the value that a policy achieves against a worst-case opponent.
The policy applies to both player 1 and player 2, and hence we have a 2-player
symmetric zero-sum game, so the game value is zero for both players, and hence
value-vs-best-response is equal to exploitability.

We construct information sets, each consisting of a list of (state, probability)
pairs where probability is a counterfactual reach probability, i.e. the
probability that the state would be reached if the best responder (the current
player) played to reach it. This is the product of the probabilities of the
necessary chance events and opponent action choices required to reach the node.

These probabilities give us the correct weighting for possible states of the
world when considering our best response for a particular information set.

The values we calculate are values of being in the specific state. Unlike in a
CFR algorithm, they are not weighted by reach probabilities. These values
take into account the whole state, so they may depend on information which is
unknown to the best-responding player.
"""

import collections

import numpy as np

from open_spiel.python.algorithms import best_response as pyspiel_best_response
import pyspiel


def _state_values(state, num_players, policy):
  """Value of a state for every player given a policy."""
  if state.is_terminal():
    return np.array(state.returns())
  else:
    p_action = (
        state.chance_outcomes() if state.is_chance_node() else
        policy.action_probabilities(state).items())
    return sum(prob * _state_values(state.child(action), num_players, policy)
               for action, prob in p_action)


def best_response(game, policy, player_id):
  """Returns information about the specified player's best response.

  Given a game and a policy for every player, computes for a single player their
  best unilateral strategy. Returns the value improvement that player would
  get, the action they should take in each information state, and the value
  of each state when following their unilateral policy.

  Args:
    game: An open_spiel game, e.g. kuhn_poker
    policy: A `policy.Policy` object. This policy should depend only on the
      information state available to the current player, but this is not
      enforced.
    player_id: The integer id of a player in the game for whom the best response
      will be computed.

  Returns:
    A dictionary of values, with keys:
      best_response_action: The best unilateral strategy for `player_id` as a
        map from infostatekey to action_id.
      best_response_state_value: The value obtained for `player_id` when
        unilaterally switching strategy, for each state.
      best_response_value: The value obtained for `player_id` when unilaterally
        switching strategy.
      info_sets: A dict of info sets, mapping info state key to a list of
        `(state, counterfactual_reach_prob)` pairs.
      nash_conv: `best_response_value - on_policy_value`
      on_policy_value: The value for `player_id` when all players follow the
        policy
      on_policy_values: The value for each player when all players follow the
        policy
  """
  root_state = game.new_initial_state()
  br = pyspiel_best_response.BestResponsePolicy(game, player_id, policy,
                                                root_state)
  on_policy_values = _state_values(root_state, game.num_players(), policy)
  best_response_value = br.value(root_state)

  # Get best response action for unvisited states
  for infostate in set(br.infosets) - set(br.cache_best_response_action):
    br.best_response_action(infostate)

  return {
      "best_response_action": br.cache_best_response_action,
      "best_response_state_value": br.cache_value,
      "best_response_value": best_response_value,
      "info_sets": br.infosets,
      "nash_conv": best_response_value - on_policy_values[player_id],
      "on_policy_value": on_policy_values[player_id],
      "on_policy_values": on_policy_values,
  }


def exploitability(game, policy):
  """Returns the exploitability of the policy in the game.

  This is implemented only for 2 players constant-sum games, and is equivalent
  to NashConv / num_players in that case. Prefer using `nash_conv`.

  Args:
    game: An open_spiel game, e.g. kuhn_poker
    policy: A `policy.Policy` object. This policy should depend only on the
      information state available to the current player, but this is not
      enforced.

  Returns:
    The value that this policy achieves when playing against the worst-case
    non-cheating opponent, averaged across both starting positions. It has a
    minimum of zero (assuming the supplied policy is non-cheating) and
    this bound is achievable in a 2p game.

  Raises:
    ValueError if the game is not a two-player constant-sum turn-based game.
  """
  if game.num_players() != 2:
    raise ValueError("Game must be a 2-player game")
  game_info = game.get_type()
  if game_info.dynamics != pyspiel.GameType.Dynamics.SEQUENTIAL:
    raise ValueError("The game must be turn-based, not {}".format(
        game_info.dynamics))
  if game_info.utility not in (pyspiel.GameType.Utility.ZERO_SUM,
                               pyspiel.GameType.Utility.CONSTANT_SUM):
    raise ValueError("The game must be constant- or zero-sum, not {}".format(
        game_info.utility))
  root_state = game.new_initial_state()
  nash_conv_value = (
      sum(
          pyspiel_best_response.CPPBestResponsePolicy(
              game, best_responder, policy).value(root_state)
          for best_responder in range(game.num_players())) - game.utility_sum())
  return nash_conv_value / game.num_players()


_NashConvReturn = collections.namedtuple("_NashConvReturn",
                                         ["nash_conv", "player_improvements"])


def nash_conv(game, policy, return_only_nash_conv=True, use_cpp_br=False):
  r"""Returns a measure of closeness to Nash for a policy in the game.

  See https://arxiv.org/pdf/1711.00832.pdf for the NashConv definition.

  Args:
    game: An open_spiel game, e.g. kuhn_poker
    policy: A `policy.Policy` object. This policy should depend only on the
      information state available to the current player, but this is not
      enforced.
    return_only_nash_conv: Whether to only return the NashConv value, or a
      namedtuple containing additional statistics. Prefer using `False`, as we
      hope to change the default to that value.
    use_cpp_br: if True, compute the best response in c++

  Returns:
    Returns a object with the following attributes:
    - player_improvements: A `[num_players]` numpy array of the improvement
      for players (i.e. value_player_p_versus_BR - value_player_p).
    - nash_conv: The sum over all players of the improvements in value that each
      player could obtain by unilaterally changing their strategy, i.e.
      sum(player_improvements).
  """
  root_state = game.new_initial_state()
  if use_cpp_br:
    best_response_values = np.array([
        pyspiel_best_response.CPPBestResponsePolicy(
            game, best_responder, policy).value(root_state)
        for best_responder in range(game.num_players())
    ])
  else:
    best_response_values = np.array([
        pyspiel_best_response.BestResponsePolicy(
            game, best_responder, policy).value(root_state)
        for best_responder in range(game.num_players())
    ])
  on_policy_values = _state_values(root_state, game.num_players(), policy)
  player_improvements = best_response_values - on_policy_values
  nash_conv_ = sum(player_improvements)
  if return_only_nash_conv:
    return nash_conv_
  else:
    return _NashConvReturn(
        nash_conv=nash_conv_, player_improvements=player_improvements)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.dqn."""

import tensorflow.compat.v1 as tf

from open_spiel.python import rl_environment
from open_spiel.python.algorithms import dqn
import pyspiel

# Temporarily disable TF2 behavior until code is updated.
tf.disable_v2_behavior()

# A simple two-action game encoded as an EFG game. Going left gets -1, going
# right gets a +1.
SIMPLE_EFG_DATA = """
  EFG 2 R "Simple single-agent problem" { "Player 1" } ""
  p "ROOT" 1 1 "ROOT" { "L" "R" } 0
    t "L" 1 "Outcome L" { -1.0 }
    t "R" 2 "Outcome R" { 1.0 }
"""


class DQNTest(tf.test.TestCase):

  def test_simple_game(self):
    game = pyspiel.load_efg_game(SIMPLE_EFG_DATA)
    env = rl_environment.Environment(game=game)
    with self.session() as sess:
      agent = dqn.DQN(sess, 0,
                      state_representation_size=
                      game.information_state_tensor_shape()[0],
                      num_actions=game.num_distinct_actions(),
                      hidden_layers_sizes=[16],
                      replay_buffer_capacity=100,
                      batch_size=5,
                      epsilon_start=0.02,
                      epsilon_end=0.01)
      total_reward = 0
      sess.run(tf.global_variables_initializer())

      for _ in range(100):
        time_step = env.reset()
        while not time_step.last():
          agent_output = agent.step(time_step)
          time_step = env.step([agent_output.action])
          total_reward += time_step.rewards[0]
        agent.step(time_step)
      self.assertGreaterEqual(total_reward, 75)

  def test_run_tic_tac_toe(self):
    env = rl_environment.Environment("tic_tac_toe")
    state_size = env.observation_spec()["info_state"][0]
    num_actions = env.action_spec()["num_actions"]

    with self.session() as sess:
      agents = [
          dqn.DQN(  # pylint: disable=g-complex-comprehension
              sess,
              player_id,
              state_representation_size=state_size,
              num_actions=num_actions,
              hidden_layers_sizes=[16],
              replay_buffer_capacity=10,
              batch_size=5) for player_id in [0, 1]
      ]
      sess.run(tf.global_variables_initializer())
      time_step = env.reset()
      while not time_step.last():
        current_player = time_step.observations["current_player"]
        current_agent = agents[current_player]
        agent_output = current_agent.step(time_step)
        time_step = env.step([agent_output.action])

      for agent in agents:
        agent.step(time_step)

  def test_run_hanabi(self):
    # Hanabi is an optional game, so check we have it before running the test.
    game = "hanabi"
    if game not in pyspiel.registered_names():
      return

    num_players = 3
    env_configs = {
        "players": num_players,
        "max_life_tokens": 1,
        "colors": 2,
        "ranks": 3,
        "hand_size": 2,
        "max_information_tokens": 3,
        "discount": 0.
    }
    env = rl_environment.Environment(game, **env_configs)
    state_size = env.observation_spec()["info_state"][0]
    num_actions = env.action_spec()["num_actions"]

    with self.session() as sess:
      agents = [
          dqn.DQN(  # pylint: disable=g-complex-comprehension
              sess,
              player_id,
              state_representation_size=state_size,
              num_actions=num_actions,
              hidden_layers_sizes=[16],
              replay_buffer_capacity=10,
              batch_size=5) for player_id in range(num_players)
      ]
      sess.run(tf.global_variables_initializer())
      time_step = env.reset()
      while not time_step.last():
        current_player = time_step.observations["current_player"]
        agent_output = [agent.step(time_step) for agent in agents]
        time_step = env.step([agent_output[current_player].action])

      for agent in agents:
        agent.step(time_step)


if __name__ == "__main__":
  tf.test.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Abstract class for meta trainers (Generalized PSRO, RNR, ...)

Meta-algorithm with modular behaviour, allowing implementation of PSRO, RNR, and
other variations.
"""

import numpy as np
from open_spiel.python.algorithms.psro_v2 import meta_strategies
from open_spiel.python.algorithms.psro_v2 import strategy_selectors
from open_spiel.python.algorithms.psro_v2 import utils

_DEFAULT_STRATEGY_SELECTION_METHOD = "probabilistic"
_DEFAULT_META_STRATEGY_METHOD = "prd"


def _process_string_or_callable(string_or_callable, dictionary):
  """Process a callable or a string representing a callable.

  Args:
    string_or_callable: Either a string or a callable
    dictionary: Dictionary of shape {string_reference: callable}

  Returns:
    string_or_callable if string_or_callable is a callable ; otherwise,
    dictionary[string_or_callable]

  Raises:
    NotImplementedError: If string_or_callable is of the wrong type, or has an
      unexpected value (Not present in dictionary).
  """
  if callable(string_or_callable):
    return string_or_callable

  try:
    return dictionary[string_or_callable]
  except KeyError:
    raise NotImplementedError("Input type / value not supported. Accepted types"
                              ": string, callable. Acceptable string values : "
                              "{}. Input provided : {}".format(
                                  list(dictionary.keys()), string_or_callable))


def sample_episode(state, policies):
  """Samples an episode using policies, starting from state.

  Args:
    state: Pyspiel state representing the current state.
    policies: List of policy representing the policy executed by each player.

  Returns:
    The result of the call to returns() of the final state in the episode.
        Meant to be a win/loss integer.
  """
  if state.is_terminal():
    return np.array(state.returns(), dtype=np.float32)

  if state.is_simultaneous_node():
    actions = [None] * state.num_players()
    for player in range(state.num_players()):
      state_policy = policies[player](state, player)
      outcomes, probs = zip(*state_policy.items())
      actions[player] = utils.random_choice(outcomes, probs)
    state.apply_actions(actions)
    return sample_episode(state, policies)

  if state.is_chance_node():
    outcomes, probs = zip(*state.chance_outcomes())
  else:
    player = state.current_player()
    state_policy = policies[player](state)
    outcomes, probs = zip(*state_policy.items())

  state.apply_action(utils.random_choice(outcomes, probs))
  return sample_episode(state, policies)


class AbstractMetaTrainer(object):
  """Abstract class implementing meta trainers.

  If a trainer is something that computes a best response to given environment &
  agents, a meta trainer will compute which best responses to compute (Against
  what, how, etc)
  This class can support PBT, Hyperparameter Evolution, etc.
  """

  # pylint:disable=dangerous-default-value
  def __init__(self,
               game,
               oracle,
               initial_policies=None,
               meta_strategy_method=_DEFAULT_META_STRATEGY_METHOD,
               training_strategy_selector=_DEFAULT_STRATEGY_SELECTION_METHOD,
               symmetric_game=False,
               number_policies_selected=1,
               **kwargs):
    """Abstract Initialization for meta trainers.

    Args:
      game: A pyspiel game object.
      oracle: An oracle object, from an implementation of the AbstractOracle
        class.
      initial_policies: A list of initial policies, to set up a default for
        training. Resorts to tabular policies if not set.
      meta_strategy_method: String, or callable taking a MetaTrainer object and
        returning a list of meta strategies (One list entry per player).
        String value can be:
              - "uniform": Uniform distribution on policies.
              - "nash": Taking nash distribution. Only works for 2 player, 0-sum
                games.
              - "prd": Projected Replicator Dynamics, as described in Lanctot et
                Al.
      training_strategy_selector: A callable or a string. If a callable, takes
        as arguments: - An instance of `PSROSolver`, - a
          `number_policies_selected` integer. and returning a list of
          `num_players` lists of selected policies to train from.
        When a string, supported values are:
              - "top_k_probabilites": selects the first
                'number_policies_selected' policies with highest selection
                probabilities.
              - "probabilistic": randomly selects 'number_policies_selected'
                with probabilities determined by the meta strategies.
              - "exhaustive": selects every policy of every player.
              - "rectified": only selects strategies that have nonzero chance of
                being selected.
              - "uniform": randomly selects 'number_policies_selected' policies
                with uniform probabilities.
      symmetric_game: Whether to consider the current game as symmetric (True)
        game or not (False).
      number_policies_selected: Maximum number of new policies to train for each
        player at each PSRO iteration.
      **kwargs: kwargs for meta strategy computation and training strategy
        selection
    """
    self._iterations = 0
    self._game = game
    self._oracle = oracle
    self._num_players = self._game.num_players()

    self.symmetric_game = symmetric_game
    self._game_num_players = self._num_players
    self._num_players = 1 if symmetric_game else self._num_players

    self._number_policies_selected = number_policies_selected

    meta_strategy_method = _process_string_or_callable(
        meta_strategy_method, meta_strategies.META_STRATEGY_METHODS)
    print("Using {} as strategy method.".format(meta_strategy_method))

    self._training_strategy_selector = _process_string_or_callable(
        training_strategy_selector,
        strategy_selectors.TRAINING_STRATEGY_SELECTORS)
    print("Using {} as training strategy selector.".format(
        self._training_strategy_selector))

    self._meta_strategy_method = meta_strategy_method
    self._kwargs = kwargs

    self._initialize_policy(initial_policies)
    self._initialize_game_state()
    self.update_meta_strategies()

  def _initialize_policy(self, initial_policies):
    return NotImplementedError(
        "initialize_policy not implemented. Initial policies passed as"
        " arguments : {}".format(initial_policies))

  def _initialize_game_state(self):
    return NotImplementedError("initialize_game_state not implemented.")

  def iteration(self, seed=None):
    """Main trainer loop.

    Args:
      seed: Seed for random BR noise generation.
    """
    self._iterations += 1
    self.update_agents()  # Generate new, Best Response agents via oracle.
    self.update_empirical_gamestate(seed=seed)  # Update gamestate matrix.
    self.update_meta_strategies()  # Compute meta strategy (e.g. Nash)

  def update_meta_strategies(self):
    self._meta_strategy_probabilities = self._meta_strategy_method(self)
    if self.symmetric_game:
      self._meta_strategy_probabilities = [self._meta_strategy_probabilities[0]]

  def update_agents(self):
    return NotImplementedError("update_agents not implemented.")

  def update_empirical_gamestate(self, seed=None):
    return NotImplementedError("update_empirical_gamestate not implemented."
                               " Seed passed as argument : {}".format(seed))

  def sample_episodes(self, policies, num_episodes):
    """Samples episodes and averages their returns.

    Args:
      policies: A list of policies representing the policies executed by each
        player.
      num_episodes: Number of episodes to execute to estimate average return of
        policies.

    Returns:
      Average episode return over num episodes.
    """
    totals = np.zeros(self._num_players)
    for _ in range(num_episodes):
      totals += sample_episode(self._game.new_initial_state(),
                               policies).reshape(-1)
    return totals / num_episodes

  def get_meta_strategies(self):
    """Returns the Nash Equilibrium distribution on meta game matrix."""
    meta_strategy_probabilities = self._meta_strategy_probabilities
    if self.symmetric_game:
      meta_strategy_probabilities = self._game_num_players * meta_strategy_probabilities
    return [np.copy(a) for a in meta_strategy_probabilities]

  def get_meta_game(self):
    """Returns the meta game matrix."""
    meta_games = self._meta_games
    if self.symmetric_game:
      meta_games = self._game_num_players * meta_games
    return [np.copy(a) for a in meta_games]

  def get_policies(self):
    """Returns the players' policies."""
    policies = self._policies
    if self.symmetric_game:
      policies = self._game_num_players * policies
    return policies

  def get_kwargs(self):
    return self._kwargs



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Symmetric payoff tensor utils."""

import itertools
import math

from absl import logging  # pylint:disable=unused-import

import numpy as np


def sym(pt):
  """Symmetrize stack of payoff tensors (stacked along first dimension).

  A payoff tensor can be `symmetrized' by averaging over all possible
  permutations of the players. This means permuting the axes corresponding to
  the player strategies as well as the payoffs assigned to the players. E.g.,
  player A playing strategy 1 and player B playing strategy 3 is no different
  from player A playing strategy 3 and player B playing strategy 1 in a
  symmetric game. Note we permuted the strategies, but we must also permute the
  payoffs.

  Args:
    pt: tensor of shape: (num_players,) + (num_strategies,) * num_players
  Returns:
    pt_sym: symmetrized payoff tensor of same shape
  """
  num_players = len(pt.shape[1:])
  num_perms = math.factorial(num_players)
  pt_sym = np.zeros_like(pt)
  logging.info('Symmetrizing over {:d} permutations...'.format(num_perms))
  for i, perm_players in enumerate(itertools.permutations(range(num_players))):
    if (i % (num_perms // 5)) == 0:
      logging.info('\t{:d} / {:d}'.format(i, num_perms))
    perm_axes = tuple([pi + 1 for pi in perm_players])
    permuted_tensor = np.transpose(pt, (0,) + perm_axes)[list(perm_players)]
    pt_sym += permuted_tensor / float(num_perms)
  logging.info('\t{total:d} / {total:d}'.format(total=num_perms))
  return pt_sym



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Run deep online mirror descent algorithm with Munchausen DQN agents."""

import os
from typing import Sequence

from absl import flags
import jax
from jax.config import config

from open_spiel.python import policy
from open_spiel.python import rl_environment
from open_spiel.python.mfg import utils
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import munchausen_deep_mirror_descent
from open_spiel.python.mfg.algorithms import nash_conv
from open_spiel.python.mfg.algorithms import policy_value
from open_spiel.python.mfg.games import factory
from open_spiel.python.utils import app
from open_spiel.python.utils import metrics

FLAGS = flags.FLAGS

flags.DEFINE_string("game_name", "mfg_crowd_modelling_2d", "Name of the game.")
flags.DEFINE_string(
    "env_setting", "crowd_modelling_2d_four_rooms",
    "Name of the game settings. If None, the game name will be used.")

# Training options.
flags.DEFINE_integer("batch_size", 128,
                     "Number of transitions to sample at each learning step.")
flags.DEFINE_integer("learn_every", 64,
                     "Number of steps between learning updates.")
flags.DEFINE_integer("num_episodes_per_iteration", 1000,
                     "Number of training eepisodes for each iteration.")
flags.DEFINE_integer("num_iterations", 100, "Number of iterations.")
flags.DEFINE_integer("epsilon_decay_duration", 100000,
                     "Number of game steps over which epsilon is decayed.")
flags.DEFINE_float("epsilon_power", 1, "Power for the epsilon decay.")
flags.DEFINE_float("epsilon_start", 0.1, "Starting exploration parameter.")
flags.DEFINE_float("epsilon_end", 0.1, "Final exploration parameter.")
flags.DEFINE_float("discount_factor", 1.0,
                   "Discount factor for future rewards.")
flags.DEFINE_bool(
    "reset_replay_buffer_on_update", False,
    "Reset the replay buffer when the softmax policy is updated.")
flags.DEFINE_integer("seed", 42, "Training seed.")
# Evaluation options.
flags.DEFINE_integer("eval_every", 200,
                     "Episode frequency at which the agents are evaluated.")
# Network options.
flags.DEFINE_list("hidden_layers_sizes", ["128", "128"],
                  "Number of hidden units in the avg-net and Q-net.")
flags.DEFINE_integer("update_target_network_every", 200,
                     "Number of steps between DQN target network updates.")
# Replay buffer options.
flags.DEFINE_integer("replay_buffer_capacity", 40000,
                     "Size of the replay buffer.")
flags.DEFINE_integer("min_buffer_size_to_learn", 1000,
                     "Number of samples in buffer before learning begins.")
# Loss and optimizer options.
flags.DEFINE_enum("optimizer", "adam", ["sgd", "adam"], "Optimizer.")
flags.DEFINE_float("learning_rate", 0.01, "Learning rate for inner rl agent.")
flags.DEFINE_enum("loss", "mse", ["mse", "huber"], "Loss function.")
flags.DEFINE_float("huber_loss_parameter", 1.0, "Parameter for Huber loss.")
flags.DEFINE_float("gradient_clipping", None, "Value to clip the gradient to.")
# Munchausen options.
flags.DEFINE_float("tau", 10, "Temperature parameter in Munchausen target.")
flags.DEFINE_float("alpha", 0.99, "Alpha parameter in Munchausen target.")
flags.DEFINE_bool("with_munchausen", True,
                  "If true, target uses Munchausen penalty terms.")
# Logging options.
flags.DEFINE_bool("use_checkpoints", False, "Save/load neural network weights.")
flags.DEFINE_string("checkpoint_dir", "/tmp/dqn_test",
                    "Directory to save/load the agent.")
flags.DEFINE_string(
    "logdir", None,
    "Logging dir to use for TF summary files. If None, the metrics will only "
    "be logged to stderr.")
flags.DEFINE_bool("log_distribution", False,
                  "Enables logging of the distribution.")


def main(argv: Sequence[str]) -> None:
  if len(argv) > 1:
    raise app.UsageError("Too many command-line arguments.")

  game = factory.create_game_with_setting(FLAGS.game_name, FLAGS.env_setting)

  num_players = game.num_players()

  # Create the environments with uniform initial policy.
  uniform_policy = policy.UniformRandomPolicy(game)
  uniform_dist = distribution.DistributionPolicy(game, uniform_policy)

  envs = [
      rl_environment.Environment(  # pylint: disable=g-complex-comprehension
          game,
          mfg_distribution=uniform_dist,
          mfg_population=p,
          observation_type=rl_environment.ObservationType.OBSERVATION)
      for p in range(num_players)
  ]

  env = envs[0]
  info_state_size = env.observation_spec()["info_state"][0]
  num_actions = env.action_spec()["num_actions"]

  # Create the agents.
  kwargs = {
      "alpha": FLAGS.alpha,
      "batch_size": FLAGS.batch_size,
      "discount_factor": FLAGS.discount_factor,
      "epsilon_decay_duration": FLAGS.epsilon_decay_duration,
      "epsilon_end": FLAGS.epsilon_end,
      "epsilon_power": FLAGS.epsilon_power,
      "epsilon_start": FLAGS.epsilon_start,
      "gradient_clipping": FLAGS.gradient_clipping,
      "hidden_layers_sizes": [int(l) for l in FLAGS.hidden_layers_sizes],
      "huber_loss_parameter": FLAGS.huber_loss_parameter,
      "learn_every": FLAGS.learn_every,
      "learning_rate": FLAGS.learning_rate,
      "loss": FLAGS.loss,
      "min_buffer_size_to_learn": FLAGS.min_buffer_size_to_learn,
      "optimizer": FLAGS.optimizer,
      "replay_buffer_capacity": FLAGS.replay_buffer_capacity,
      "reset_replay_buffer_on_update": FLAGS.reset_replay_buffer_on_update,
      "seed": FLAGS.seed,
      "tau": FLAGS.tau,
      "update_target_network_every": FLAGS.update_target_network_every,
      "with_munchausen": FLAGS.with_munchausen
  }
  agents = [
      munchausen_deep_mirror_descent.MunchausenDQN(p, info_state_size,
                                                   num_actions, **kwargs)
      for p in range(num_players)
  ]

  # Metrics writer will also log the metrics to stderr.
  just_logging = FLAGS.logdir is None or jax.host_id() > 0
  writer = metrics.create_default_writer(
      logdir=FLAGS.logdir, just_logging=just_logging)

  # # Save the parameters.
  writer.write_hparams(kwargs)

  def logging_fn(it, episode, vals):
    writer.write_scalars(it * num_episodes_per_iteration + episode, vals)

  num_episodes_per_iteration = FLAGS.num_episodes_per_iteration
  md = munchausen_deep_mirror_descent.DeepOnlineMirrorDescent(
      game,
      envs,
      agents,
      eval_every=FLAGS.eval_every,
      num_episodes_per_iteration=num_episodes_per_iteration,
      logging_fn=logging_fn)

  def log_metrics(it):
    """Logs the training metrics for each iteration."""
    initial_states = game.new_initial_states()
    pi_value = policy_value.PolicyValue(game, md.distribution, md.policy)
    m = {
        f"best_response/{state}": pi_value.eval_state(state)
        for state in initial_states
    }
    nash_conv_md = nash_conv.NashConv(game, md.policy).nash_conv()
    m["nash_conv_md"] = nash_conv_md
    if FLAGS.log_distribution and FLAGS.logdir:
      # We log distribution directly to a Pickle file as it may be large for
      # logging as a metric.
      filename = os.path.join(FLAGS.logdir, f"distribution_{it}.pkl")
      utils.save_parametric_distribution(md.distribution, filename)
    logging_fn(it, 0, m)

  log_metrics(0)
  for it in range(1, FLAGS.num_iterations + 1):
    md.iteration()
    log_metrics(it)

  # Make sure all values were written.
  writer.flush()


if __name__ == "__main__":
  config.parse_flags_with_absl()
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mean field games implemented in Python.

These games are registered as they are imported. It's perfectly possible to
import just a single game if you prefer. There is no need to add new games here,
so long as they register themselves and you import them when wanting to use
them. However, adding them here will make them available for playthroughs and
for automated API testing.

Registration looks like this:
```
pyspiel.register_game(_GAME_TYPE, KuhnPokerGame)
```
"""
from open_spiel.python.mfg.games import crowd_modelling
from open_spiel.python.mfg.games import dynamic_routing
from open_spiel.python.mfg.games import linear_quadratic
from open_spiel.python.mfg.games import predator_prey



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Utils module for dynamic routing game and mean field routing game.

This module has three main classes:
- Network
- Vehicle
- OriginDestinationDemand
"""

from typing import Any, Dict, Iterable, List, Optional, Tuple

# In case one vehicle has reached a end node, then it cannot do anything. In
# this case its action is 0. Action 0 is reserved to encode no possible action
# as requested by Open Spiel.
NO_POSSIBLE_ACTION = 0


def _nodes_to_road_section(origin: str, destination: str) -> str:
  """Create a road section 'A->B' from two nodes 'A' and 'B'."""
  return f"{origin}->{destination}"


def _road_section_to_nodes(movement: str) -> Tuple[str, str]:
  """Split a road section 'A->B' to two nodes 'A' and 'B'."""
  origin, destination = movement.split("->")
  return origin, destination


def assign_dictionary_input_to_object(dict_object: Dict[str, Any],
                                      road_sections: Iterable[str],
                                      default_value: Any) -> Dict[str, Any]:
  """Check dictionary has road sections has key or return default_value dict."""
  if dict_object:
    assert set(dict_object) == set(road_sections), (
        "Objects are not defined for each road sections.")
    return dict_object
  dict_object_returned = {}
  for road_section in road_sections:
    dict_object_returned[road_section] = default_value
  return dict_object_returned


class Network:
  """Network implementation.

  A network is basically a directed graph with a volume delay function on each
  of its edges. Each vertex is refered to as a string (for example "A") and each
  edge as a string f"{node1}->{node2}" (for example "A->B"). The network is
  created from a adjacency list. Each road section is mapped to an action index
  (positive integer) in _road_section_to_action, and vice versa in
  _action_to_road_section. The volume delay function on each road section rs is
  given by _free_flow_travel_time[rs]*(1+ _a[rs]*(v/_capacity[rs])**_b[rs])
  where v is the volume on the road section rs, according to the U.S. Bureau of
  Public Road (BPR). Such functions are called fundamental diagram of traffic
  flow.

  If one would like to plot the network then node position should be passed
  in the constructor. Then return_list_for_matplotlib_quiver can be used with
  Matplotlib:
  ```python3
  fig, ax = plt.subplots()
  o_xs, o_ys, d_xs, d_ys = g.return_list_for_matplotlib_quiver()
  ax.quiver(o_xs, o_ys, np.subtract(d_xs, o_xs), np.subtract(d_ys, o_ys),
            color="b", angles='xy', scale_units='xy', scale=1)
  ```

  See the Network tests for an example.
  Attributes:
    _a, _b, _capacity, _free_flow_travel_time: dictionary that maps road section
      string representation to its a, b, relative capacity and free flow travel
      time coefficient in its BPR function.
    _action_to_road_section: dictionary that maps action id to road section.
    _adjacency_list: adjacency list of the line graph of the road network.
    _node_position: dictionary that maps node to couple of float encoding x and
      y position of the node. None by default.
    _road_section_to_action: dictionary that maps road section to action id.
  """
  _a: Dict[str, float]
  _b: Dict[str, float]
  _adjacency_list: Dict[str, Iterable[str]]
  _capacity: Dict[str, float]
  _free_flow_travel_time: Dict[str, float]
  _node_position: Dict[str, Tuple[float, float]]
  _road_section_to_action: Dict[str, int]

  def __init__(self,
               adjacency_list: Dict[str, Iterable[str]],
               node_position: Optional[Dict[str, Tuple[float, float]]] = None,
               bpr_a_coefficient: Optional[Dict[str, float]] = None,
               bpr_b_coefficient: Optional[Dict[str, float]] = None,
               capacity: Optional[Dict[str, float]] = None,
               free_flow_travel_time: Optional[Dict[str, float]] = None):
    self._adjacency_list = adjacency_list
    self._road_section_to_action, self._action_to_road_section = (
        self._create_movement_to_action_and_action_to_road_section())

    nodes = set(adjacency_list)
    # pylint: disable=g-complex-comprehension
    assert all(destination_node in nodes
               for destination_nodes in self._adjacency_list.values()
               for destination_node in destination_nodes), (
                   "Adjacency list is not correct.")
    # pylint: enable=g-complex-comprehension
    if node_position:
      assert set(node_position) == nodes
      self._node_position = node_position
    else:
      self._node_position = None
    self._a = assign_dictionary_input_to_object(bpr_a_coefficient,
                                                self._road_section_to_action, 0)
    self._b = assign_dictionary_input_to_object(bpr_b_coefficient,
                                                self._road_section_to_action, 1)
    self._capacity = assign_dictionary_input_to_object(
        capacity, self._road_section_to_action, 1)
    self._free_flow_travel_time = assign_dictionary_input_to_object(
        free_flow_travel_time, self._road_section_to_action, 1)
    assert hasattr(self, "_adjacency_list")
    assert hasattr(self, "_node_position")
    assert hasattr(self, "_a")
    assert hasattr(self, "_b")
    assert hasattr(self, "_capacity")
    assert hasattr(self, "_free_flow_travel_time")

  def _create_movement_to_action_and_action_to_road_section(
      self) -> Tuple[Dict[str, int], Dict[int, str]]:
    """Create dictionary that maps movement to action.

    The dictionary that maps movement to action is used to define the action
    from a movement that a vehicle would like to do. The dictionary that maps an
    action to the destintion of the movement is used to move a vehicle that does
    an action to the destination of its movement.
    Returns:
      road_section_to_action: dictionary with key begin a movement for example
        "O->A" and value the action numbers. Action numbers are succesive
        integers indexed from 1.
      action_to_road_section: map an action number to the end node of the
        movement. if road_section_to_action["O->A"] = 0 then,
        action_to_road_section[0] = "O->A"
    """
    road_section_to_action = {}
    action_to_road_section = {}
    action_number = 1
    for origin, successors in self._adjacency_list.items():
      for destination in successors:
        road_section = _nodes_to_road_section(origin, destination)
        if road_section in road_section_to_action:
          raise ValueError((
              f"{road_section} exists twice in the adjacency list. The current "
              "network implementation does not enable parallel links."))
        road_section_to_action[road_section] = action_number
        action_to_road_section[action_number] = road_section
        action_number += 1
    return road_section_to_action, action_to_road_section

  def num_links(self) -> int:
    """Returns the number of road sections."""
    return len(self._road_section_to_action)

  def num_actions(self) -> int:
    """Returns the number of possible actions.

    Equal to the number of road section + 1. An action could either be moving to
    a specific road section or not move.
    """
    return 1 + self.num_links()

  def links(self) -> List[str]:
    """Returns the road sections as a list."""
    return list(self._road_section_to_action)

  def get_successors(self, node: str) -> Iterable[str]:
    """Returns the successor nodes of the node."""
    return self._adjacency_list[node]

  def get_action_id_from_movement(self, origin: str, destination: str) -> int:
    """Maps two connected nodes to an action."""
    return self._road_section_to_action[_nodes_to_road_section(
        origin, destination)]

  def get_road_section_from_action_id(self, action_id: int) -> str:
    """Maps a action to the corresponding road section."""
    return self._action_to_road_section[action_id]

  def is_location_at_sink_node(self, road_section: str) -> bool:
    """Returns True if the road section has no successors."""
    start_section, end_section_node = _road_section_to_nodes(road_section)
    if start_section not in self._adjacency_list:
      raise KeyError(f"{start_section} is not a network node.")
    return not self.get_successors(end_section_node)

  def check_list_of_vehicles_is_correct(self, vehicles: Iterable["Vehicle"]):
    """Assert that vehicles have valid origin and destination."""
    for vehicle in vehicles:
      if (vehicle.origin not in self._road_section_to_action or
          vehicle.destination not in self._road_section_to_action):
        raise ValueError(f"Incorrect origin or destination for {vehicle}")

  def check_list_of_od_demand_is_correct(
      self, vehicles: Iterable["OriginDestinationDemand"]):
    """Assert that OD demands have valid origin and destination."""
    for vehicle in vehicles:
      if (vehicle.origin not in self._road_section_to_action or
          vehicle.destination not in self._road_section_to_action):
        raise ValueError(f"Incorrect origin or destination for {vehicle}")

  def __str__(self) -> str:
    return str(self._adjacency_list)

  def get_travel_time(self, road_section: str, volume: float) -> int:
    """Returns travel time on the road section given the volume on it.

    Volume unit should be the same as the capacity unit.
    Travel time unit is the free flow travel time unit.
    Args:
      road_section: the road section.
      volume: the volume on the road section.
    """
    return self._free_flow_travel_time[road_section] * (
        1.0 + self._a[road_section] *
        (volume / self._capacity[road_section])**self._b[road_section])

  def assert_valid_action(self, action: int, road_section: str = None):
    """Assert that an action as a int is valid.

    The action should be a int between 1 and num_actions. In case road_section
    is not None then it is test if the action correspond to going on a road
    section which is a successor of road_section.

    Args:
      action: the action,
      road_section: the road section.
    """
    assert isinstance(action, int), f"{action} is not a int."
    assert 1 <= action < self.num_actions(), str(action)
    if road_section is not None:
      new_road_section = self.get_road_section_from_action_id(action)
      origin_new_section, end_new_section = _road_section_to_nodes(
          new_road_section)
      _, end_section_node = _road_section_to_nodes(road_section)
      assert end_section_node == origin_new_section, (
          f"The action is not legal, trying to go to {new_road_section} "
          f"from {road_section} without going through {end_section_node}"
          ".")
      successors = self.get_successors(origin_new_section)
      assert end_new_section in successors, (
          f"Invalid action {new_road_section}. It is not a successors of"
          f" {end_section_node}: {successors}.")

  def return_position_of_road_section(self,
                                      road_section: str) -> Tuple[float, float]:
    """Returns position of the middle of theroad section as (x,y)."""
    assert self._node_position is not None, (
        "The network should have node positions in order to be plot.")
    o_link, d_link = _road_section_to_nodes(road_section)
    o_x, o_y = self._node_position[o_link]
    d_x, d_y = self._node_position[d_link]
    return (o_x + d_x) / 2, (o_y + d_y) / 2

  def return_list_for_matplotlib_quiver(
      self) -> Tuple[List[float], List[float], List[float], List[float]]:
    """Returns 4 list of encoding the positions of the road sections.

    ```python3
    fig, ax = plt.subplots()
    o_xs, o_ys, d_xs, d_ys = g.return_list_for_matplotlib_quiver()
    ax.quiver(o_xs, o_ys, np.subtract(d_xs, o_xs), np.subtract(d_ys, o_ys),
              color="b", angles='xy', scale_units='xy', scale=1)
    ```
    will show the network.
    Returns:
      o_xs, o_ys, d_xs, d_ys: list of the start x and y positions and of the end
        x and y postions of each road section. Each element of each list
        corresponds to one road section.
    """
    assert self._node_position is not None, (
        "The network should have node positions in order to be plot.")
    o_xs = []
    o_ys = []
    d_xs = []
    d_ys = []
    for road_section in self._road_section_to_action:
      o_link, d_link = _road_section_to_nodes(road_section)
      o_x, o_y = self._node_position[o_link]
      d_x, d_y = self._node_position[d_link]
      o_xs.append(o_x)
      o_ys.append(o_y)
      d_xs.append(d_x)
      d_ys.append(d_y)
    return o_xs, o_ys, d_xs, d_ys


class Vehicle:
  """A Vehicle is one origin and one destination.

  Both the origin and the destination of the vehicle are road section, therefore
  they are string formatted as "{str}->{str}".
  Attributes:
    destination: destination of the vehicle.
    origin: origin of the vehicle.
    departure_time: departure time of the vehicle.
  """
  _destination: str
  _origin: str
  _departure_time: float

  def __init__(self,
               origin: str,
               destination: str,
               departure_time: float = 0.0):
    assert all("->" in node for node in [origin, destination])
    self._origin = origin
    self._destination = destination
    self._departure_time = departure_time

  @property
  def origin(self) -> str:
    """Returns vehicle's origin."""
    return self._origin

  @property
  def destination(self) -> str:
    """Returns vehicle's destination."""
    return self._destination

  @property
  def departure_time(self) -> float:
    """Returns vehicle's departure time."""
    return self._departure_time

  def __str__(self):
    return (f"Vehicle with origin {self.origin}, destination {self.destination}"
            f" and departure time {self._departure_time}.")


class OriginDestinationDemand(Vehicle):
  """Number of trips from origin to destination for a specific departure time.

  Both the origin and the destination of the vehicle are road section, therefore
  they are string formatted as "{str}->{str}".
  Attributes:
    destination: destination of the vehicles.
    origin: origin of the vehicles.
    departure_time: departure time of the vehicles.
    counts: the number of vehicles with the origin, destination and departure
      time.
  """
  _counts: float

  def __init__(self, origin: str, destination: str, departure_time: float,
               counts: float):
    super().__init__(origin, destination, departure_time)
    self._counts = counts

  @property
  def counts(self) -> float:
    """Returns the number of vehicles in the instance."""
    return self._counts

  def __str__(self):
    return (f"{self._counts} with origin {self.origin}, destination "
            f"{self.destination} and departure time {self._departure_time}.")



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An Oracle for Exact Best Responses.

This class computes the best responses against sets of policies.
"""

from open_spiel.python import policy as openspiel_policy
from open_spiel.python.algorithms import best_response
from open_spiel.python.algorithms import policy_utils
from open_spiel.python.algorithms.psro_v2 import optimization_oracle
from open_spiel.python.algorithms.psro_v2 import utils
import pyspiel


class BestResponseOracle(optimization_oracle.AbstractOracle):
  """Oracle using exact best responses to compute BR to policies."""

  def __init__(self,
               best_response_backend='cpp',
               game=None,
               all_states=None,
               state_to_information_state=None,
               **kwargs):
    """Init function for the RLOracle.

    Args:
      best_response_backend: A string (either 'cpp' or 'py'), specifying the
        best response backend to use (C++ or python, respectively). The cpp
        backend should be preferred, generally, as it is significantly faster.
      game: The game on which the optimization process takes place.
      all_states: The result of calling get_all_states.get_all_states. Cached
        for improved performance.
      state_to_information_state: A dict mapping str(state) to
        state.information_state for every state in the game. Cached for improved
        performance.
      **kwargs: kwargs
    """
    super(BestResponseOracle, self).__init__(**kwargs)
    self.best_response_backend = best_response_backend
    if self.best_response_backend == 'cpp':
      # Should compute all_states and state_to_information_state only once in
      # the program, as caching them speeds up TabularBestResponse tremendously.
      self.all_states, self.state_to_information_state = (
          utils.compute_states_and_info_states_if_none(
              game, all_states, state_to_information_state))

      policy = openspiel_policy.UniformRandomPolicy(game)

      policy_to_dict = policy_utils.policy_to_dict(
          policy, game, self.all_states, self.state_to_information_state)

      # pylint: disable=g-complex-comprehension
      # Cache TabularBestResponse for players, due to their costly construction
      # TODO(b/140426861): Use a single best-responder once the code supports
      # multiple player ids.
      self.best_response_processors = [
          pyspiel.TabularBestResponse(game, best_responder_id, policy_to_dict)
          for best_responder_id in range(game.num_players())
      ]
      self.best_responders = [
          best_response.CPPBestResponsePolicy(
              game, i_player, policy, self.all_states,
              self.state_to_information_state,
              self.best_response_processors[i_player]
          )
          for i_player in range(game.num_players())
      ]
      # pylint: enable=g-complex-comprehension

  def __call__(self,
               game,
               training_parameters,
               strategy_sampler=utils.sample_strategy,
               using_joint_strategies=False,
               **oracle_specific_execution_kwargs):
    """Call method for oracle, returns best responses for training_parameters.

    Args:
      game: The game on which the optimization process takes place.
      training_parameters: List of list of dicts: one list per player, one dict
        per selected agent in the pool for each player,
        each dictionary containing the following fields:
        - policy: the policy from which to start training.
        - total_policies: A list of all policy.Policy strategies used for
          training, including the one for the current player. Either
          marginalized or joint strategies are accepted.
        - current_player: Integer representing the current player.
        - probabilities_of_playing_policies: A list of arrays representing, per
          player, the probabilities of playing each policy in total_policies for
          the same player.
      strategy_sampler: Callable that samples strategies from `total_policies`
        using `probabilities_of_playing_policies`. It only samples one joint
        "action" for all players. Implemented to be able to take into account
        joint probabilities of action.
      using_joint_strategies: Whether the meta-strategies sent are joint (True)
        or marginalized.
      **oracle_specific_execution_kwargs: Other set of arguments, for
        compatibility purposes. Can for example represent whether to Rectify
        Training or not.

    Returns:
      A list of list of OpenSpiel Policy objects representing the expected
      best response, following the same structure as training_parameters.
    """
    new_policies = []
    for player_parameters in training_parameters:
      player_policies = []
      for params in player_parameters:
        current_player = params['current_player']
        total_policies = params['total_policies']
        probabilities_of_playing_policies = params[
            'probabilities_of_playing_policies']
        if using_joint_strategies:
          aggr_policy = utils.aggregate_joint_policies(
              game, utils.marginal_to_joint(total_policies),
              probabilities_of_playing_policies.reshape(-1))
        else:
          aggr_policy = utils.aggregate_policies(
              game, total_policies, probabilities_of_playing_policies)

        # This takes as input an aggregate policy, and computes a best response
        # for current_player at the applicable information states by recursing
        # through the game tree. At information states involving other players
        # or chance, the aggr_policy is used to compute the expected value, such
        # that a best response for current_player can be computed.
        if self.best_response_backend == 'py':
          best_resp = best_response.BestResponsePolicy(game, current_player,
                                                       aggr_policy)
        else:
          self.best_response_processors[current_player].set_policy(
              policy_utils.policy_to_dict(aggr_policy, game, self.all_states,
                                          self.state_to_information_state))

          self.best_responders[current_player] = (
              best_response.CPPBestResponsePolicy(
                  game, current_player, aggr_policy, self.all_states,
                  self.state_to_information_state,
                  self.best_response_processors[current_player]))
          best_resp = self.best_responders[current_player]
        player_policies.append(best_resp)
      new_policies.append(player_policies)
    return new_policies



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Implementation of a mean field routing game.

The game is derived from https://arxiv.org/abs/2110.11943.
It is the extension of the dynamic routing game python_dynamic_routing.
The list of vehicles decribing the N player of the dynamic routing game is
replaced by a list of OriginDestinationDemand. One OriginDestinationDemand
corresponds to one population of vehicles (with the same origin, destination and
departure time).

This game is a variant of the mean field route choice game as the vehicle
movement depends on the current network congestion. In the mean field route
choice game, the number of time step to reach the destination is constant and
does not depend on the network congestion, neither of the vehicle cost function.
In the dynamic driving and routing games the vehicle choose its speed to travel
on each link in order to minimize its cost function. Therefore the congestion is
encoded in the cost function.

More context can be found on the docstring of the python_dynamic_routing class.
"""
import functools
from typing import Any, Iterable, List, Mapping, Optional, Tuple

import numpy as np

from open_spiel.python.games import dynamic_routing_data
from open_spiel.python.games import dynamic_routing_utils
from open_spiel.python.observation import IIGObserverForPublicInfoGame
import pyspiel

_DEFAULT_PARAMS = {
    "max_num_time_step": 10,
    "time_step_length": 0.5,
    "players": -1
}
_GAME_TYPE = pyspiel.GameType(
    short_name="python_mfg_dynamic_routing",
    long_name="Python Mean Field Routing Game",
    dynamics=pyspiel.GameType.Dynamics.MEAN_FIELD,
    chance_mode=pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC,
    information=pyspiel.GameType.Information.PERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.GENERAL_SUM,
    reward_model=pyspiel.GameType.RewardModel.REWARDS,
    max_num_players=1,
    min_num_players=1,
    provides_information_state_string=True,
    provides_information_state_tensor=True,
    provides_observation_string=True,
    provides_observation_tensor=True,
    default_loadable=True,
    provides_factored_observation_string=True,
    parameter_specification=_DEFAULT_PARAMS)

WAITING_TIME_NOT_ASSIGNED = -1


@functools.lru_cache(maxsize=None)
def _state_to_str(
    is_chance_init: bool,
    location: str,
    time_step: int,
    player_id: int,
    waiting_time: int,
    destination: str,
    final_travel_time: float,
) -> str:
  """Convert the state to a string representation.

  As the string representation will be used in dictionaries for various
  algorithms that computes the state value, expected return, best response or
  find the mean field Nash equilibrium.
  The state is uniquely define by the current time, the type of node
  (decision, mean field or chance), the vehicle location, its destination and
  its waiting time.
  Args:
    is_chance_init: True if at chance initialization.
    location: the location of the representative player.
    time_step: the current time step.
    player_id: the current node type as a player id.
    waiting_time: the representative player waiting time.
    destination: the destination of the representative player.
    final_travel_time: time of arrival.

  Returns:
    state_string: string representing uniquely the mean field game.
  """
  if is_chance_init:
    return "initial chance node"
  if player_id == pyspiel.PlayerId.DEFAULT_PLAYER_ID:
    time = str(time_step)
  elif player_id == pyspiel.PlayerId.MEAN_FIELD:
    time = f"{time_step}_mean_field"
  elif player_id == pyspiel.PlayerId.CHANCE:
    time = f"{time_step}_chance"
  else:
    raise ValueError(
        "Player id should be DEFAULT_PLAYER_ID, MEAN_FIELD or CHANCE")
  if final_travel_time:
    return (f"Arrived at {location}, with travel time "
            f"{final_travel_time}, t={time}")
  return (f"Location={location}, waiting_time={waiting_time},"
          f" t={time}, destination='{destination}'")


class MeanFieldRoutingGame(pyspiel.Game):
  """Implementation of mean field routing game.

  The representative vehicle/player is represented as a tuple current location,
  current waiting time and destination. When the waiting time is negative, the
  vehicle choose on with successor link it would like to go. When arriving on
  the link, a waiting time is assigned to the player based on the distribution
  of players on the link. The vehicle travel time is equal to the time step when
  they first reach their destination. See module docstring for more information.

  Attributes inherited from GameInfo:
    max_chance_outcomes: maximum number of chance actions. Set to the length of
      od_demand, i.e. the number of `OriginDestinationDemand`s.
    max_game_length: maximum number of time step played. Passed during
      construction.
    max_utility: maximum utility is the opposite of the minimum travel
      time. Set to 0.
    min_utility: minimum utility is the opposite of the maximum travel
      time. Set to - max_game_length - 1.
    num_distinct_actions: maximum number of possible actions. This is
      equal to the number of links + 1 (corresponding to having no
      possible action _NO_POSSIBLE_ACTION).
    num_players: the number of vehicles. Should be 1 as this mean field
      game is a one population game.
  Attributes:
    network: the network of the game.
    od_demand: a list of the vehicle. Their origin and their destination should
      be road sections of the game.
    time_step_length: size of the time step, used to convert travel times into
      number of game time steps.
    perform_sanity_checks: if true, sanity checks are done during the game,
      should be set to false to speed up the game.
    total_num_vehicle: total number of vehicles as the sum of the od_demand.
    chance_outcomes: chance outcomes based on the initial probability
      distribution and their probabilities.
  """
  network: dynamic_routing_utils.Network
  od_demand: List[dynamic_routing_utils.OriginDestinationDemand]
  perform_sanity_checks: bool
  time_step_length: float

  def __init__(self,
               params: Mapping[str, Any],
               network: Optional[dynamic_routing_utils.Network] = None,
               od_demand: Optional[List[
                   dynamic_routing_utils.OriginDestinationDemand]] = None,
               perform_sanity_checks: bool = True):
    """Initiliaze the game.

    Args:
      params: game parameters. It should define max_num_time_step and
        time_step_length.
      network: the network of the game.
      od_demand: a list of the vehicle. Their origin and their destination
        should be road sections of the game.
      perform_sanity_checks: set the perform_sanity_checks attribute.
    """
    max_num_time_step = params["max_num_time_step"]
    time_step_length = params["time_step_length"]
    self.network = network if network else dynamic_routing_data.BRAESS_NETWORK
    self.od_demand = (
        od_demand
        if od_demand else dynamic_routing_data.BRAESS_NETWORK_OD_DEMAND)
    self.network.check_list_of_od_demand_is_correct(self.od_demand)
    self.perform_sanity_checks = perform_sanity_checks
    self.time_step_length = time_step_length
    self.total_num_vehicle = sum(
        [od_demand_item.counts for od_demand_item in self.od_demand])
    self.chance_outcomes = [(i, od_demand_item.counts / self.total_num_vehicle)
                            for i, od_demand_item in enumerate(self.od_demand)]
    game_info = pyspiel.GameInfo(
        num_distinct_actions=self.network.num_actions(),
        max_chance_outcomes=len(self.od_demand),
        num_players=1,
        min_utility=-max_num_time_step - 1,
        max_utility=0,
        max_game_length=max_num_time_step)
    super().__init__(_GAME_TYPE, game_info, params if params else {})

  def new_initial_state(self) -> "MeanFieldRoutingGameState":
    """Returns the state corresponding to the start of a game."""
    return MeanFieldRoutingGameState(self, self.time_step_length)

  def make_py_observer(self, iig_obs_type=None, params=None):
    """Returns a NetworkObserver object used for observing game state."""
    if ((iig_obs_type is None) or
        (iig_obs_type.public_info and not iig_obs_type.perfect_recall)):
      return NetworkObserver(self.network.num_actions(), self.max_game_length())
    return IIGObserverForPublicInfoGame(iig_obs_type, params)

  def max_chance_nodes_in_history(self):
    """Maximun chance nodes in game history."""
    return self.max_game_length() + 1

  def get_road_section_as_int(self, section: Optional[str]) -> int:
    """Returns the integer representation of the road section."""
    if section is None:
      return 0
    start_node, end_node = (
        dynamic_routing_utils._road_section_to_nodes(section))  # pylint:disable=protected-access
    return self.network.get_action_id_from_movement(start_node, end_node)


class MeanFieldRoutingGameState(pyspiel.State):
  """State of the DynamicRoutingGame.

  One player is equal to one vehicle.
  See docstring of the game class and of the file for more information.
  Attributes:
    _current_time_step: current time step of the game.
    _is_chance_init: boolean that encodes weither the current node is the
      initial chance node.
    _is_terminal: boolean that encodes weither the game is over.
    _max_travel_time: int that encodes maximum travel time on any link in number
      of time steps. Needed to enumerate all the possible state of a vehicle
      being on a link to compute volume of cars on the link.
    _max_waiting_time: maximum time a vehicle can wait on a time. This is done
      in order to limit the number of possible state with a vehicle on a
      specific link.
    _normed_density_on_vehicle_link: density of vehicles on the link that is
      used by the representative vehicle. This is given by the mean field
      distribution.
    _time_step_length: size of the time step, used to convert travel times into
      number of game time steps.
    _vehicle_at_destination: boolean that encodes if the representative vehicle
      has reached its destination.
    _vehicle_destination: the destination of the representative vehicle
      corresponding to this state. It is associated to the representative
      vehicle after the initial chance node according to the od_demand
      distribution.
    _vehicle_final_travel_time: the travel time of the representative vehicle,
      the travel is either 0 if the vehicle is still in the network or its
      travel time if the vehicle has reached its destination.
    _vehicle_location: current location of the vehicle as a network road
      section.
    _vehicle_without_legal_action: boolean that encodes if the representative
      vehicle has reach a sink node, meaning that it will not be able to move
      anymore.
    _waiting_time: time that the vehicle has to wait before moving to the next
      link (equal to the link travel time when the vehicle just reached the
      link).
  """
  _current_time_step: int
  _is_chance_init: bool
  _is_terminal: bool
  _max_travel_time: int
  _max_waiting_time: int
  _normed_density_on_vehicle_link: float
  _time_step_length: float
  _vehicle_at_destination: bool
  _vehicle_destination: Optional[str]
  _vehicle_final_travel_time: float
  _vehicle_location: Optional[str]
  _vehicle_without_legal_action: bool
  _waiting_time: int

  def __init__(self, game: MeanFieldRoutingGame, time_step_length: float):
    """Constructor; should only be called by Game.new_initial_state."""
    super().__init__(game)
    self._current_time_step = 0
    self._is_chance_init = True  # is true for the first state of the game.
    self._is_terminal = False
    if self.get_game().perform_sanity_checks:
      assert game.num_players() == 1, (
          "This mean field routing game should have a unique player.")
    self._player_id = pyspiel.PlayerId.CHANCE
    self._time_step_length = time_step_length
    self._vehicle_at_destination = False
    self._vehicle_final_travel_time = 0.0
    self._vehicle_without_legal_action = False
    self._vehicle_location = None
    self._vehicle_destination = None
    self._max_travel_time = self.get_game().max_game_length()
    # TODO(cabannes): cap maximum link waiting time to faster simulations.
    self._max_waiting_time = self._max_travel_time
    self._waiting_time = WAITING_TIME_NOT_ASSIGNED

  @property
  def current_time_step(self) -> int:
    """Return current time step."""
    return self._current_time_step

  def current_player(self) -> pyspiel.PlayerId:
    """Returns the current player."""
    if self._is_terminal:
      return pyspiel.PlayerId.TERMINAL
    return self._player_id

  def state_to_str(self,
                   location: str,
                   time_step: int,
                   player_id: int = pyspiel.PlayerId.DEFAULT_PLAYER_ID,
                   waiting_time: int = 0,
                   destination: str = ""):
    """Convert the state to a string representation."""
    return _state_to_str(
        self._is_chance_init,
        location,
        time_step,
        player_id,
        waiting_time,
        destination or self._vehicle_destination,
        self._vehicle_final_travel_time,
    )

  def distribution_support(self) -> List[str]:
    """Returns the state that should be used for update_distribution.

    The distribution of the vehicle is used to determined the number of
    cars on the same link of the representative vehicle in order to define
    the waiting time of the representative vehicle when joining a link.
    Therefore, only the states corresponding to be on the link of the
    representative vehicle at this current time are useful.
    Returns:
      list of the two state: being on the link of the representative vehicle at
        the current time and being stuck in traffic or not.
    """
    if self._vehicle_without_legal_action:
      return []
    od_demand = self.get_game().od_demand
    dist = [
        self.state_to_str(  # pylint:disable=g-complex-comprehension
            self._vehicle_location,
            self._current_time_step,
            player_id=pyspiel.PlayerId.MEAN_FIELD,
            waiting_time=waiting_time,
            destination=destination)
        for waiting_time in range(WAITING_TIME_NOT_ASSIGNED,
                                  self._max_travel_time)
        for destination in {od._destination for od in od_demand}  # pylint:disable=protected-access
    ]
    assert len(set(dist)) == len(dist), (
        f"Distribution should not have duplicated states: {dist}.")
    return dist

  def update_distribution(self, distribution: List[float]):
    """Get the number of cars on the same link as the representative player.

    _normed_density_on_vehicle_link stores the number of cars on the link
    where the representative player is.
    Args:
      distribution: the probability for a vehicle to be in the states in
        distribution_support. The distribution is a list of probabilities.
    """
    game = self.get_game()
    if game.perform_sanity_checks:
      if self._player_id != pyspiel.PlayerId.MEAN_FIELD:
        raise ValueError(("update_distribution should only be called at"
                          " a MEAN_FIELD state."))
    self._player_id = pyspiel.PlayerId.DEFAULT_PLAYER_ID
    if not self._vehicle_without_legal_action:
      self._normed_density_on_vehicle_link = sum(distribution)
      if game.perform_sanity_checks:
        assert 0 <= self._normed_density_on_vehicle_link <= 1 + 1e-4, (
            f"{self._normed_density_on_vehicle_link} is not in [0, 1].")
      if self._waiting_time == WAITING_TIME_NOT_ASSIGNED:
        volume = (game.total_num_vehicle * self._normed_density_on_vehicle_link)
        self._waiting_time = int(
            game.network.get_travel_time(self._vehicle_location, volume) /
            self._time_step_length) - 1
        self._waiting_time = max(0, self._waiting_time)

  def chance_outcomes(self) -> List[Tuple[int, float]]:
    """Returns the initial probability distribution is returned.

    One chance outcome correspond to each possible OD pair with a departure
    time, the probability of each chance outcome is the proportion of vehicle in
    each OD pair with a departure time.
    Returns:
      list_tuple_outcome_probabilities: chance outcomes and their probability.
    """
    game = self.get_game()
    if game.perform_sanity_checks:
      assert self._player_id == pyspiel.PlayerId.CHANCE
      assert self._is_chance_init
    return game.chance_outcomes

  def _legal_actions(self, player: pyspiel.PlayerId) -> List[int]:
    """Return the legal actions of the vehicle.

    Legal actions are the succesor road section of the vehicle current road
    section.
    Args:
      player: the vehicle id.

    Returns:
      list_legal_actions: a list of legal actions. If the game is finished then
        the list is empty. If the vehicle is at its destination, has a positive
        waiting time or if it is on a node without successors then an empty list
        is returned. Otherwise the list of successors nodes of the current
        vehicle location is returned.
    """
    if self._is_terminal:
      return []
    if self.get_game().perform_sanity_checks:
      assert player == pyspiel.PlayerId.DEFAULT_PLAYER_ID, str(player)
    if self._vehicle_without_legal_action:
      # If the vehicle is at destination it cannot do anything.
      return [dynamic_routing_utils.NO_POSSIBLE_ACTION]
    if self._waiting_time > 0:
      return [dynamic_routing_utils.NO_POSSIBLE_ACTION]
    _, end_section_node = dynamic_routing_utils._road_section_to_nodes(  # pylint:disable=protected-access
        self._vehicle_location)
    successors = self.get_game().network.get_successors(end_section_node)
    if self.get_game().perform_sanity_checks:
      if not successors:
        raise ValueError(("If a vehicle is not without legal action, it"
                          " should have an action."))
      assert isinstance(successors, Iterable)
    actions = [
        self.get_game().network.get_action_id_from_movement(
            end_section_node, d) for d in successors
    ]
    map(self.get_game().network.assert_valid_action, actions)
    return sorted(actions)

  def _apply_action(self, action: int):
    """Apply the action to the state.

    This function can be either called on a chance node or on a decision
    node. If called on the initial chance node, the action gives in which OD
    demand the representative vehicle belongs too (it put the vehicle at
    this location and define its destination).
    If called on decision node, the action defines on which link the vehicle
    will move (if it is not stuck in traffic) and assign a waiting time to the
    vehicle.
    Args:
      action: the action to apply.
    """
    if self._player_id == pyspiel.PlayerId.CHANCE:
      self._player_id = pyspiel.PlayerId.DEFAULT_PLAYER_ID
      assert self._is_chance_init
      # Apply action is called on initial chance node to initialized
      # the vehicle position based on the initial location
      # distribution.
      od_demand = self.get_game().od_demand
      self._vehicle_destination = od_demand[action].destination
      self._vehicle_location = od_demand[action].origin
      self._waiting_time = int(od_demand[action].departure_time /
                               self._time_step_length)
      self._is_chance_init = False
      self._normed_density_on_vehicle_link = 0
    elif self._player_id == pyspiel.PlayerId.DEFAULT_PLAYER_ID:
      self._player_id = pyspiel.PlayerId.MEAN_FIELD
      # Apply action is called on a descision node. If the vehicle can
      # move, then it will move to the next road section.
      # Has the vehicle already reached a sink node?
      if not self._vehicle_without_legal_action:
        # If the vehicle is stuck in traffic it cannot move.
        if self._waiting_time > 0:
          self._waiting_time -= 1
        else:
          if self.get_game().perform_sanity_checks:
            self.get_game().network.assert_valid_action(action,
                                                        self._vehicle_location)
          self._vehicle_location = (
              self.get_game().network.get_road_section_from_action_id(action))
          # Has the vehicle just reached its destination?
          if self._vehicle_location == self._vehicle_destination:
            self._vehicle_final_travel_time = self._current_time_step
            self._vehicle_at_destination = True
            self._vehicle_without_legal_action = True
          # Will the vehicle have a legal action for next time step?
          elif self.get_game().network.is_location_at_sink_node(
              self._vehicle_location):
            self._vehicle_without_legal_action = True
            self._vehicle_final_travel_time = -self.get_game().min_utility()
          else:
            self._waiting_time = WAITING_TIME_NOT_ASSIGNED
      self._current_time_step += 1
    elif self.get_game().perform_sanity_checks:
      if self._is_terminal:
        raise ValueError(
            "_apply_action should not be called at a end of the game.")
      if self._player_id == pyspiel.PlayerId.MEAN_FIELD:
        raise ValueError(
            "_apply_action should not be called at a MEAN_FIELD state.")
    # Is the game finished?
    if self._current_time_step >= self.get_game().max_game_length():
      self._is_terminal = True
      if not self._vehicle_at_destination:
        self._vehicle_final_travel_time = -self.get_game().min_utility()

  def _action_to_string(self, player, action) -> str:
    """Action -> string."""
    if player == pyspiel.PlayerId.CHANCE:
      if self._is_chance_init:
        return f"Vehicle is assigned to population {action}."
      return f"Change node; the vehicle movement is {bool(action)}."
    if self.get_game().perform_sanity_checks:
      assert player == pyspiel.PlayerId.DEFAULT_PLAYER_ID
    if action == dynamic_routing_utils.NO_POSSIBLE_ACTION:
      return f"Vehicle {player} reach a sink node or its destination."
    if self.get_game().perform_sanity_checks:
      self.get_game().network.assert_valid_action(action)
    return (f"Vehicle {player} would like to move to " + str(
        self.get_game().network.get_road_section_from_action_id(action)) + ".")

  def is_terminal(self) -> bool:
    """Returns True if the game is over."""
    return self._is_terminal

  def is_waiting(self) -> bool:
    """Returns True if the wait time is non-zero."""
    return self._waiting_time > 0

  def returns(self) -> List[float]:
    """Total reward for each player over the course of the game so far."""
    if not self._is_terminal:
      return [0]
    return [-self._vehicle_final_travel_time * self._time_step_length]

  def get_location_as_int(self) -> int:
    """Returns the vehicle location.

    This will be 1-based action index of the location, or 0 when the location is
    None before the initial chance node.
    """
    return self.get_game().get_road_section_as_int(self._vehicle_location)

  def get_destination_as_int(self) -> int:
    """Returns the vehicle destination.


    This will be 1-based action index of the destination, or 0 when the
    destination is None before the initial chance node.
    """
    return self.get_game().get_road_section_as_int(self._vehicle_destination)

  def __str__(self) -> str:
    """String for debug purposes. No particular semantics are required."""
    if self._vehicle_location is not None:
      return self.state_to_str(
          self._vehicle_location,
          self._current_time_step,
          player_id=self._player_id,
          waiting_time=self._waiting_time)
    assert self._current_time_step == 0
    return "Before initial chance node"


class NetworkObserver:
  """Network observer used by the learning algorithm.

  The state string is the state history string. The state tensor is an array
  of size number of locations * 2 + maximum number of time steps + 2, which is
  the concatenation of one-hot encodings of the location, destination (1-based;
  if location or destination is None, then the 0th element will be set to 1) and
  the current time (0-based). The last element of the array will be set to 1 if
  waiting time is positive, or 0 otherwise.

  Attributes:
    dict: Dictionary of tensors for the components of the observation
      corresponding to the location, destination and time.
    tensor: The concatenated form of the observation.
  """

  def __init__(self, num_locations: int, max_num_time_step: int):
    """Initializes an empty observation tensor."""
    self.tensor = np.zeros(num_locations * 2 + max_num_time_step + 1 + 1,
                           np.float32)
    self.dict = {
        "location": self.tensor[:num_locations],
        "destination": self.tensor[num_locations:num_locations * 2],
        "time": self.tensor[num_locations * 2:-1],
        "waiting": self.tensor[-1:]
    }

  def set_from(self, state, player):
    """Sets the state tensor based on the specified state.

    Note that the function may be called with arbitrary states of the game, e.g.
    from different runs, and therefore the tensor should be cleared and updated
    instead of preserving any earlier values.

    Args:
      state: state of the game.
      player: player id that should play.
    """
    assert player == pyspiel.PlayerId.DEFAULT_PLAYER_ID
    self.tensor.fill(0)
    self.dict["location"][state.get_location_as_int()] = 1
    self.dict["destination"][state.get_destination_as_int()] = 1
    self.dict["time"][state.current_time_step] = 1
    self.dict["waiting"][0] = state.is_waiting()

  def string_from(self, state, player):
    """Return the state history string."""
    assert player == pyspiel.PlayerId.DEFAULT_PLAYER_ID
    return str(state)


# Register the game with the OpenSpiel library
pyspiel.register_game(_GAME_TYPE, MeanFieldRoutingGame)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for Python mean field routing game."""

from absl.testing import absltest
import numpy as np
import numpy.testing as npt

from open_spiel.python import games  # pylint:disable=unused-import
from open_spiel.python import policy
from open_spiel.python.games import dynamic_routing_utils
from open_spiel.python.mfg import games as mfg_games  # pylint:disable=unused-import
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import mirror_descent
from open_spiel.python.mfg.algorithms import nash_conv
from open_spiel.python.mfg.algorithms import policy_value
from open_spiel.python.mfg.games import dynamic_routing
from open_spiel.python.mfg.games import factory
from open_spiel.python.observation import make_observation
import pyspiel

_NUMBER_OF_ITERATIONS_TESTS = 1


class MeanFieldRoutingGameTest(absltest.TestCase):
  """Checks we can create the game and clone states."""

  def test_load(self):
    """Test load and game creation."""
    game = pyspiel.load_game("python_mfg_dynamic_routing")
    game.new_initial_state()

  def test_create(self):
    """Checks we can create the game and clone states."""
    game = pyspiel.load_game("python_mfg_dynamic_routing")
    self.assertEqual(game.get_type().dynamics,
                     pyspiel.GameType.Dynamics.MEAN_FIELD)
    state = game.new_initial_state()
    state.clone()

  def test_random_game(self):
    """Test random simulation."""
    game = pyspiel.load_game("python_mfg_dynamic_routing")
    pyspiel.random_sim_test(game, num_sims=10, serialize=False, verbose=True)

  def test_evolving_trajectory_with_uniform_policy(self):
    """Test evolving distribution."""
    game = pyspiel.load_game("python_mfg_dynamic_routing")
    distribution.DistributionPolicy(game, policy.UniformRandomPolicy(game))

  def test_non_default_param_from_string(self):
    """Check params can be given through string loading."""
    game = pyspiel.load_game("python_mfg_dynamic_routing(max_num_time_step=5)")
    self.assertEqual(game.max_game_length(), 5)

  def test_non_default_param_from_dict(self):
    """Check params can be given through a dictionary."""
    game = pyspiel.load_game("python_mfg_dynamic_routing",
                             {"max_num_time_step": 5})
    self.assertEqual(game.max_game_length(), 5)

  # TODO(open_spiel): enable ficticious_play with game where the dynamics depend
  # on the distribution.
  # def test_ficticious_play(self):
  #   """Test that ficticious play can be used on this game."""
  #   mfg_game = pyspiel.load_game("python_mfg_dynamic_routing")
  #   fp = fictitious_play.FictitiousPlay(mfg_game)
  #   for _ in range(_NUMBER_OF_ITERATIONS_TESTS):
  #     fp.iteration()
  #   nash_conv.NashConv(mfg_game, fp.get_policy())

  def test_online_mirror_descent(self):
    """Test that online mirror descent can be used on this game."""
    mfg_game = pyspiel.load_game("python_mfg_dynamic_routing")
    omd = mirror_descent.MirrorDescent(mfg_game)
    for _ in range(_NUMBER_OF_ITERATIONS_TESTS):
      omd.iteration()
    nash_conv.NashConv(mfg_game, omd.get_policy())

  def test_online_mirror_descent_convergence(self):
    """Test that online mirror descent converges to equilibrium in default game."""
    mfg_game = pyspiel.load_game("python_mfg_dynamic_routing", {
        "time_step_length": 0.05,
        "max_num_time_step": 100
    })
    omd = mirror_descent.MirrorDescent(mfg_game, lr=1)
    for _ in range(50):
      omd.iteration()
    self.assertAlmostEqual(
        nash_conv.NashConv(mfg_game, omd.get_policy()).nash_conv(), 0)

  def test_braess_paradox(self):
    """Test that Braess paradox can be reproduced with the mean field game."""
    mfg_game = pyspiel.load_game("python_mfg_dynamic_routing", {
        "time_step_length": 0.05,
        "max_num_time_step": 100
    })

    class NashEquilibriumBraess(policy.Policy):

      def action_probabilities(self, state, player_id=None):
        legal_actions = state.legal_actions()
        if not legal_actions:
          return {dynamic_routing_utils.NO_POSSIBLE_ACTION: 1.0}
        elif len(legal_actions) == 1:
          return {legal_actions[0]: 1.0}
        else:
          if legal_actions[0] == 2:
            return {2: 0.75, 3: 0.25}
          elif legal_actions[0] == 4:
            return {4: 2 / 3, 5: 1 / 3}
        raise ValueError(f"{legal_actions} is not correct.")

    ne_policy = NashEquilibriumBraess(mfg_game, 1)
    self.assertEqual(
        -policy_value.PolicyValue(
            mfg_game, distribution.DistributionPolicy(mfg_game, ne_policy),
            ne_policy).value(mfg_game.new_initial_state()), 3.75)
    self.assertEqual(nash_conv.NashConv(mfg_game, ne_policy).nash_conv(), 0.0)

    class SocialOptimumBraess(policy.Policy):

      def action_probabilities(self, state, player_id=None):
        legal_actions = state.legal_actions()
        if not legal_actions:
          return {dynamic_routing_utils.NO_POSSIBLE_ACTION: 1.0}
        elif len(legal_actions) == 1:
          return {legal_actions[0]: 1.0}
        else:
          if legal_actions[0] == 2:
            return {2: 0.5, 3: 0.5}
          elif legal_actions[0] == 4:
            return {5: 1.0}
        raise ValueError(f"{legal_actions} is not correct.")

    so_policy = SocialOptimumBraess(mfg_game, 1)
    self.assertEqual(
        -policy_value.PolicyValue(
            mfg_game, distribution.DistributionPolicy(mfg_game, so_policy),
            so_policy).value(mfg_game.new_initial_state()), 3.5)
    self.assertEqual(nash_conv.NashConv(mfg_game, so_policy).nash_conv(), 0.75)

  def test_vehicle_origin_outside_network(self):
    """Check raise assertion if vehicle's origin is outside the Network."""
    od_demand = [
        dynamic_routing_utils.OriginDestinationDemand("I->O", "D->E", 0, 5)
    ]
    with self.assertRaises(ValueError):
      dynamic_routing.MeanFieldRoutingGame(
          {
              "max_num_time_step": 10,
              "time_step_length": 0.5,
              "players": -1
          },
          od_demand=od_demand)

  def test_vehicle_destination_outside_network(self):
    """Check raise assertion if vehicle's destination is outside the Network."""
    od_demand = [
        dynamic_routing_utils.OriginDestinationDemand("O->A", "E->F", 0, 5)
    ]
    with self.assertRaises(ValueError):
      dynamic_routing.MeanFieldRoutingGame(
          {
              "max_num_time_step": 10,
              "time_step_length": 0.5,
              "players": -1
          },
          od_demand=od_demand)

  def test_multiple_departure_time_vehicle(self):
    """Check that departure time can be define."""
    od_demand = [
        dynamic_routing_utils.OriginDestinationDemand("O->A", "D->E", 0, 5),
        dynamic_routing_utils.OriginDestinationDemand("O->A", "D->E", 0.5, 5),
        dynamic_routing_utils.OriginDestinationDemand("O->A", "D->E", 1.0, 5)
    ]
    game = dynamic_routing.MeanFieldRoutingGame(
        {
            "max_num_time_step": 10,
            "time_step_length": 0.5,
            "players": -1
        },
        od_demand=od_demand)
    pyspiel.random_sim_test(game, num_sims=10, serialize=False, verbose=True)

  def test_game_evolution_uniform_policy(self):
    """Check game evolution under uniform policy."""
    # TODO(cabannes): test evolution of the game as expected (test value of the
    # state).
    # TODO(cabannes): test legal_actions().

  def test_observer_correct(self):
    """Checks that the observer is correctly updated."""
    game = pyspiel.load_game("python_mfg_dynamic_routing")
    num_locations, steps = 8, 10
    self.assertEqual(game.num_distinct_actions(), num_locations)
    self.assertEqual(game.max_game_length(), steps)
    py_obs = make_observation(game)

    state = game.new_initial_state()
    self.assertEqual(state.current_player(), pyspiel.PlayerId.CHANCE)

    state.apply_action(0)
    self.assertEqual(state.current_player(), 0)

    location, destination = 1, 7
    self.assertEqual(state.get_location_as_int(), location)
    self.assertEqual(state.get_destination_as_int(), destination)

    py_obs.set_from(state, state.current_player())
    obs_size = num_locations * 2 + steps + 2
    expected_tensor = np.zeros(obs_size)
    # location = 1
    # destination + num_locations = 15
    # time + 2 * num_locations = 16
    # waiting bit at last index.
    expected_tensor[[1, 15, 16]] = 1
    npt.assert_array_equal(py_obs.tensor, expected_tensor)

  def test_apply_actions_error_no_movement_with_negative_waiting_time(self):
    """Check that a vehicle cannot choose to not move if it has to move."""
    # TODO(cabannes): test apply_actions().

  def test_apply_actions_error_wrong_movement_with_negative_waiting_time(self):
    """Check that a vehicle cannot choose to move to a not successor link."""
    # TODO(cabannes): test apply_actions().

  def test_apply_actions_error_movement_with_positive_waiting_time(self):
    """Check that a vehicle cannot choose to move if it cannot move yet."""
    # TODO(cabannes): test apply_actions().

  @absltest.skip(
      "Test of OMD on Sioux Falls is disabled as it takes a long time to run.")
  def test_online_mirror_descent_sioux_falls_dummy(self):
    """Test that online mirror descent can be used on the Sioux Falls game."""
    mfg_game = factory.create_game_with_setting(
        "python_mfg_dynamic_routing",
        "dynamic_routing_sioux_falls_dummy_demand")
    omd = mirror_descent.MirrorDescent(mfg_game)
    for _ in range(_NUMBER_OF_ITERATIONS_TESTS):
      omd.iteration()
    nash_conv.NashConv(mfg_game, omd.get_policy())


if __name__ == "__main__":
  absltest.main()
